# Copyright 2015 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from .i18n import register_translation


# Based on https://pkg-isocodes.alioth.debian.org/
# which is licensed under LGPL-2.1+
# Generated using _print_iso_639() (see below)
_ISO_639 = [
    ("Afar", "aa", "aar", ""),
    ("Abkhazian", "ab", "abk", ""),
    ("Achinese", "", "ace", ""),
    ("Acoli", "", "ach", ""),
    ("Adangme", "", "ada", ""),
    ("Adyghe; Adygei", "", "ady", ""),
    ("Afro-Asiatic languages", "", "afa", ""),
    ("Afrihili", "", "afh", ""),
    ("Afrikaans", "af", "afr", ""),
    ("Ainu", "", "ain", ""),
    ("Akan", "ak", "aka", ""),
    ("Akkadian", "", "akk", ""),
    ("Albanian", "sq", "alb", "sqi"),
    ("Aleut", "", "ale", ""),
    ("Algonquian languages", "", "alg", ""),
    ("Southern Altai", "", "alt", ""),
    ("Amharic", "am", "amh", ""),
    ("English, Old (ca. 450-1100)", "", "ang", ""),
    ("Angika", "", "anp", ""),
    ("Apache languages", "", "apa", ""),
    ("Arabic", "ar", "ara", ""),
    ("Official Aramaic (700-300 BCE); Imperial Aramaic (700-300 BCE)", "", "arc", ""),
    ("Aragonese", "an", "arg", ""),
    ("Armenian", "hy", "arm", "hye"),
    ("Mapudungun; Mapuche", "", "arn", ""),
    ("Arapaho", "", "arp", ""),
    ("Artificial languages", "", "art", ""),
    ("Arawak", "", "arw", ""),
    ("Assamese", "as", "asm", ""),
    ("Asturian; Bable; Leonese; Asturleonese", "", "ast", ""),
    ("Athapascan languages", "", "ath", ""),
    ("Australian languages", "", "aus", ""),
    ("Avaric", "av", "ava", ""),
    ("Avestan", "ae", "ave", ""),
    ("Awadhi", "", "awa", ""),
    ("Aymara", "ay", "aym", ""),
    ("Azerbaijani", "az", "aze", ""),
    ("Banda languages", "", "bad", ""),
    ("Bamileke languages", "", "bai", ""),
    ("Bashkir", "ba", "bak", ""),
    ("Baluchi", "", "bal", ""),
    ("Bambara", "bm", "bam", ""),
    ("Balinese", "", "ban", ""),
    ("Basque", "eu", "baq", "eus"),
    ("Basa", "", "bas", ""),
    ("Baltic languages", "", "bat", ""),
    ("Beja; Bedawiyet", "", "bej", ""),
    ("Belarusian", "be", "bel", ""),
    ("Bemba", "", "bem", ""),
    ("Bengali", "bn", "ben", ""),
    ("Berber languages", "", "ber", ""),
    ("Bhojpuri", "", "bho", ""),
    ("Bihari languages", "bh", "bih", ""),
    ("Bikol", "", "bik", ""),
    ("Bini; Edo", "", "bin", ""),
    ("Bislama", "bi", "bis", ""),
    ("Siksika", "", "bla", ""),
    ("Bantu languages", "", "bnt", ""),
    ("Bosnian", "bs", "bos", ""),
    ("Braj", "", "bra", ""),
    ("Breton", "br", "bre", ""),
    ("Batak languages", "", "btk", ""),
    ("Buriat", "", "bua", ""),
    ("Buginese", "", "bug", ""),
    ("Bulgarian", "bg", "bul", ""),
    ("Burmese", "my", "bur", "mya"),
    ("Blin; Bilin", "", "byn", ""),
    ("Caddo", "", "cad", ""),
    ("Central American Indian languages", "", "cai", ""),
    ("Galibi Carib", "", "car", ""),
    ("Catalan; Valencian", "ca", "cat", ""),
    ("Caucasian languages", "", "cau", ""),
    ("Cebuano", "", "ceb", ""),
    ("Celtic languages", "", "cel", ""),
    ("Chamorro", "ch", "cha", ""),
    ("Chibcha", "", "chb", ""),
    ("Chechen", "ce", "che", ""),
    ("Chagatai", "", "chg", ""),
    ("Chinese", "zh", "chi", "zho"),
    ("Chuukese", "", "chk", ""),
    ("Mari", "", "chm", ""),
    ("Chinook jargon", "", "chn", ""),
    ("Choctaw", "", "cho", ""),
    ("Chipewyan; Dene Suline", "", "chp", ""),
    ("Cherokee", "", "chr", ""),
    (
        "Church Slavic; Old Slavonic; Church Slavonic; Old Bulgarian; "
        "Old Church Slavonic",
        "cu",
        "chu",
        "",
    ),
    ("Chuvash", "cv", "chv", ""),
    ("Cheyenne", "", "chy", ""),
    ("Chamic languages", "", "cmc", ""),
    ("Coptic", "", "cop", ""),
    ("Cornish", "kw", "cor", ""),
    ("Corsican", "co", "cos", ""),
    ("Creoles and pidgins, English based", "", "cpe", ""),
    ("Creoles and pidgins, French-based", "", "cpf", ""),
    ("Creoles and pidgins, Portuguese-based", "", "cpp", ""),
    ("Cree", "cr", "cre", ""),
    ("Crimean Tatar; Crimean Turkish", "", "crh", ""),
    ("Creoles and pidgins", "", "crp", ""),
    ("Kashubian", "", "csb", ""),
    ("Cushitic languages", "", "cus", ""),
    ("Czech", "cs", "cze", "ces"),
    ("Dakota", "", "dak", ""),
    ("Danish", "da", "dan", ""),
    ("Dargwa", "", "dar", ""),
    ("Land Dayak languages", "", "day", ""),
    ("Delaware", "", "del", ""),
    ("Slave (Athapascan)", "", "den", ""),
    ("Dogrib", "", "dgr", ""),
    ("Dinka", "", "din", ""),
    ("Divehi; Dhivehi; Maldivian", "dv", "div", ""),
    ("Dogri", "", "doi", ""),
    ("Dravidian languages", "", "dra", ""),
    ("Lower Sorbian", "", "dsb", ""),
    ("Duala", "", "dua", ""),
    ("Dutch, Middle (ca. 1050-1350)", "", "dum", ""),
    ("Dutch; Flemish", "nl", "dut", "nld"),
    ("Dyula", "", "dyu", ""),
    ("Dzongkha", "dz", "dzo", ""),
    ("Efik", "", "efi", ""),
    ("Egyptian (Ancient)", "", "egy", ""),
    ("Ekajuk", "", "eka", ""),
    ("Elamite", "", "elx", ""),
    ("English", "en", "eng", ""),
    ("English, Middle (1100-1500)", "", "enm", ""),
    ("Esperanto", "eo", "epo", ""),
    ("Estonian", "et", "est", ""),
    ("Ewe", "ee", "ewe", ""),
    ("Ewondo", "", "ewo", ""),
    ("Fang", "", "fan", ""),
    ("Faroese", "fo", "fao", ""),
    ("Fanti", "", "fat", ""),
    ("Fijian", "fj", "fij", ""),
    ("Filipino; Pilipino", "", "fil", ""),
    ("Finnish", "fi", "fin", ""),
    ("Finno-Ugrian languages", "", "fiu", ""),
    ("Fon", "", "fon", ""),
    ("French", "fr", "fre", "fra"),
    ("French, Middle (ca. 1400-1600)", "", "frm", ""),
    ("French, Old (842-ca. 1400)", "", "fro", ""),
    ("Northern Frisian", "", "frr", ""),
    ("Eastern Frisian", "", "frs", ""),
    ("Western Frisian", "fy", "fry", ""),
    ("Fulah", "ff", "ful", ""),
    ("Friulian", "", "fur", ""),
    ("Ga", "", "gaa", ""),
    ("Gayo", "", "gay", ""),
    ("Gbaya", "", "gba", ""),
    ("Germanic languages", "", "gem", ""),
    ("Georgian", "ka", "geo", "kat"),
    ("German", "de", "ger", "deu"),
    ("Geez", "", "gez", ""),
    ("Gilbertese", "", "gil", ""),
    ("Gaelic; Scottish Gaelic", "gd", "gla", ""),
    ("Irish", "ga", "gle", ""),
    ("Galician", "gl", "glg", ""),
    ("Manx", "gv", "glv", ""),
    ("German, Middle High (ca. 1050-1500)", "", "gmh", ""),
    ("German, Old High (ca. 750-1050)", "", "goh", ""),
    ("Gondi", "", "gon", ""),
    ("Gorontalo", "", "gor", ""),
    ("Gothic", "", "got", ""),
    ("Grebo", "", "grb", ""),
    ("Greek, Ancient (to 1453)", "", "grc", ""),
    ("Greek, Modern (1453-)", "el", "gre", "ell"),
    ("Guarani", "gn", "grn", ""),
    ("Swiss German; Alemannic; Alsatian", "", "gsw", ""),
    ("Gujarati", "gu", "guj", ""),
    ("Gwich'in", "", "gwi", ""),
    ("Haida", "", "hai", ""),
    ("Haitian; Haitian Creole", "ht", "hat", ""),
    ("Hausa", "ha", "hau", ""),
    ("Hawaiian", "", "haw", ""),
    ("Hebrew", "he", "heb", ""),
    ("Herero", "hz", "her", ""),
    ("Hiligaynon", "", "hil", ""),
    ("Himachali languages; Western Pahari languages", "", "him", ""),
    ("Hindi", "hi", "hin", ""),
    ("Hittite", "", "hit", ""),
    ("Hmong; Mong", "", "hmn", ""),
    ("Hiri Motu", "ho", "hmo", ""),
    ("Croatian", "hr", "hrv", ""),
    ("Upper Sorbian", "", "hsb", ""),
    ("Hungarian", "hu", "hun", ""),
    ("Hupa", "", "hup", ""),
    ("Iban", "", "iba", ""),
    ("Igbo", "ig", "ibo", ""),
    ("Icelandic", "is", "ice", "isl"),
    ("Ido", "io", "ido", ""),
    ("Sichuan Yi; Nuosu", "ii", "iii", ""),
    ("Ijo languages", "", "ijo", ""),
    ("Inuktitut", "iu", "iku", ""),
    ("Interlingue; Occidental", "ie", "ile", ""),
    ("Iloko", "", "ilo", ""),
    ("Interlingua (International Auxiliary Language Association)", "ia", "ina", ""),
    ("Indic languages", "", "inc", ""),
    ("Indonesian", "id", "ind", ""),
    ("Indo-European languages", "", "ine", ""),
    ("Ingush", "", "inh", ""),
    ("Inupiaq", "ik", "ipk", ""),
    ("Iranian languages", "", "ira", ""),
    ("Iroquoian languages", "", "iro", ""),
    ("Italian", "it", "ita", ""),
    ("Javanese", "jv", "jav", ""),
    ("Lojban", "", "jbo", ""),
    ("Japanese", "ja", "jpn", ""),
    ("Judeo-Persian", "", "jpr", ""),
    ("Judeo-Arabic", "", "jrb", ""),
    ("Kara-Kalpak", "", "kaa", ""),
    ("Kabyle", "", "kab", ""),
    ("Kachin; Jingpho", "", "kac", ""),
    ("Kalaallisut; Greenlandic", "kl", "kal", ""),
    ("Kamba", "", "kam", ""),
    ("Kannada", "kn", "kan", ""),
    ("Karen languages", "", "kar", ""),
    ("Kashmiri", "ks", "kas", ""),
    ("Kanuri", "kr", "kau", ""),
    ("Kawi", "", "kaw", ""),
    ("Kazakh", "kk", "kaz", ""),
    ("Kabardian", "", "kbd", ""),
    ("Khasi", "", "kha", ""),
    ("Khoisan languages", "", "khi", ""),
    ("Central Khmer", "km", "khm", ""),
    ("Khotanese;Sakan", "", "kho", ""),
    ("Kikuyu; Gikuyu", "ki", "kik", ""),
    ("Kinyarwanda", "rw", "kin", ""),
    ("Kirghiz; Kyrgyz", "ky", "kir", ""),
    ("Kimbundu", "", "kmb", ""),
    ("Konkani", "", "kok", ""),
    ("Komi", "kv", "kom", ""),
    ("Kongo", "kg", "kon", ""),
    ("Korean", "ko", "kor", ""),
    ("Kosraean", "", "kos", ""),
    ("Kpelle", "", "kpe", ""),
    ("Karachay-Balkar", "", "krc", ""),
    ("Karelian", "", "krl", ""),
    ("Kru languages", "", "kro", ""),
    ("Kurukh", "", "kru", ""),
    ("Kuanyama; Kwanyama", "kj", "kua", ""),
    ("Kumyk", "", "kum", ""),
    ("Kurdish", "ku", "kur", ""),
    ("Kutenai", "", "kut", ""),
    ("Ladino", "", "lad", ""),
    ("Lahnda", "", "lah", ""),
    ("Lamba", "", "lam", ""),
    ("Lao", "lo", "lao", ""),
    ("Latin", "la", "lat", ""),
    ("Latvian", "lv", "lav", ""),
    ("Lezghian", "", "lez", ""),
    ("Limburgan; Limburger; Limburgish", "li", "lim", ""),
    ("Lingala", "ln", "lin", ""),
    ("Lithuanian", "lt", "lit", ""),
    ("Mongo", "", "lol", ""),
    ("Lozi", "", "loz", ""),
    ("Luxembourgish; Letzeburgesch", "lb", "ltz", ""),
    ("Luba-Lulua", "", "lua", ""),
    ("Luba-Katanga", "lu", "lub", ""),
    ("Ganda", "lg", "lug", ""),
    ("Luiseno", "", "lui", ""),
    ("Lunda", "", "lun", ""),
    ("Luo (Kenya and Tanzania)", "", "luo", ""),
    ("Lushai", "", "lus", ""),
    ("Macedonian", "mk", "mac", "mkd"),
    ("Madurese", "", "mad", ""),
    ("Magahi", "", "mag", ""),
    ("Marshallese", "mh", "mah", ""),
    ("Maithili", "", "mai", ""),
    ("Makasar", "", "mak", ""),
    ("Malayalam", "ml", "mal", ""),
    ("Mandingo", "", "man", ""),
    ("Maori", "mi", "mao", "mri"),
    ("Austronesian languages", "", "map", ""),
    ("Marathi", "mr", "mar", ""),
    ("Masai", "", "mas", ""),
    ("Malay", "ms", "may", "msa"),
    ("Moksha", "", "mdf", ""),
    ("Mandar", "", "mdr", ""),
    ("Mende", "", "men", ""),
    ("Irish, Middle (900-1200)", "", "mga", ""),
    ("Mi'kmaq; Micmac", "", "mic", ""),
    ("Minangkabau", "", "min", ""),
    ("Uncoded languages", "", "mis", ""),
    ("Mon-Khmer languages", "", "mkh", ""),
    ("Malagasy", "mg", "mlg", ""),
    ("Maltese", "mt", "mlt", ""),
    ("Manchu", "", "mnc", ""),
    ("Manipuri", "", "mni", ""),
    ("Manobo languages", "", "mno", ""),
    ("Mohawk", "", "moh", ""),
    ("Moldavian; Moldovan", "mo", "mol", ""),
    ("Mongolian", "mn", "mon", ""),
    ("Mossi", "", "mos", ""),
    ("Multiple languages", "", "mul", ""),
    ("Munda languages", "", "mun", ""),
    ("Creek", "", "mus", ""),
    ("Mirandese", "", "mwl", ""),
    ("Marwari", "", "mwr", ""),
    ("Mayan languages", "", "myn", ""),
    ("Erzya", "", "myv", ""),
    ("Nahuatl languages", "", "nah", ""),
    ("North American Indian languages", "", "nai", ""),
    ("Neapolitan", "", "nap", ""),
    ("Nauru", "na", "nau", ""),
    ("Navajo; Navaho", "nv", "nav", ""),
    ("Ndebele, South; South Ndebele", "nr", "nbl", ""),
    ("Ndebele, North; North Ndebele", "nd", "nde", ""),
    ("Ndonga", "ng", "ndo", ""),
    ("Low German; Low Saxon; German, Low; Saxon, Low", "", "nds", ""),
    ("Nepali", "ne", "nep", ""),
    ("Nepal Bhasa; Newari", "", "new", ""),
    ("Nias", "", "nia", ""),
    ("Niger-Kordofanian languages", "", "nic", ""),
    ("Niuean", "", "niu", ""),
    ("Norwegian Nynorsk; Nynorsk, Norwegian", "nn", "nno", ""),
    ("Bokm\xe5l, Norwegian; Norwegian Bokm\xe5l", "nb", "nob", ""),
    ("Nogai", "", "nog", ""),
    ("Norse, Old", "", "non", ""),
    ("Norwegian", "no", "nor", ""),
    ("N'Ko", "", "nqo", ""),
    ("Pedi; Sepedi; Northern Sotho", "", "nso", ""),
    ("Nubian languages", "", "nub", ""),
    ("Classical Newari; Old Newari; Classical Nepal Bhasa", "", "nwc", ""),
    ("Chichewa; Chewa; Nyanja", "ny", "nya", ""),
    ("Nyamwezi", "", "nym", ""),
    ("Nyankole", "", "nyn", ""),
    ("Nyoro", "", "nyo", ""),
    ("Nzima", "", "nzi", ""),
    ("Occitan (post 1500)", "oc", "oci", ""),
    ("Ojibwa", "oj", "oji", ""),
    ("Oriya", "or", "ori", ""),
    ("Oromo", "om", "orm", ""),
    ("Osage", "", "osa", ""),
    ("Ossetian; Ossetic", "os", "oss", ""),
    ("Turkish, Ottoman (1500-1928)", "", "ota", ""),
    ("Otomian languages", "", "oto", ""),
    ("Papuan languages", "", "paa", ""),
    ("Pangasinan", "", "pag", ""),
    ("Pahlavi", "", "pal", ""),
    ("Pampanga; Kapampangan", "", "pam", ""),
    ("Panjabi; Punjabi", "pa", "pan", ""),
    ("Papiamento", "", "pap", ""),
    ("Palauan", "", "pau", ""),
    ("Persian, Old (ca. 600-400 B.C.)", "", "peo", ""),
    ("Persian", "fa", "per", "fas"),
    ("Philippine languages", "", "phi", ""),
    ("Phoenician", "", "phn", ""),
    ("Pali", "pi", "pli", ""),
    ("Polish", "pl", "pol", ""),
    ("Pohnpeian", "", "pon", ""),
    ("Portuguese", "pt", "por", ""),
    ("Prakrit languages", "", "pra", ""),
    ("Proven\xe7al, Old (to 1500); Occitan, Old (to 1500)", "", "pro", ""),
    ("Pushto; Pashto", "ps", "pus", ""),
    ("Reserved for local use", "", "qaa-qtz", ""),
    ("Quechua", "qu", "que", ""),
    ("Rajasthani", "", "raj", ""),
    ("Rapanui", "", "rap", ""),
    ("Rarotongan; Cook Islands Maori", "", "rar", ""),
    ("Romance languages", "", "roa", ""),
    ("Romansh", "rm", "roh", ""),
    ("Romany", "", "rom", ""),
    ("Romanian", "ro", "rum", "ron"),
    ("Rundi", "rn", "run", ""),
    ("Aromanian; Arumanian; Macedo-Romanian", "", "rup", ""),
    ("Russian", "ru", "rus", ""),
    ("Sandawe", "", "sad", ""),
    ("Sango", "sg", "sag", ""),
    ("Yakut", "", "sah", ""),
    ("South American Indian languages", "", "sai", ""),
    ("Salishan languages", "", "sal", ""),
    ("Samaritan Aramaic", "", "sam", ""),
    ("Sanskrit", "sa", "san", ""),
    ("Sasak", "", "sas", ""),
    ("Santali", "", "sat", ""),
    ("Sicilian", "", "scn", ""),
    ("Scots", "", "sco", ""),
    ("Selkup", "", "sel", ""),
    ("Semitic languages", "", "sem", ""),
    ("Irish, Old (to 900)", "", "sga", ""),
    ("Sign Languages", "", "sgn", ""),
    ("Shan", "", "shn", ""),
    ("Sidamo", "", "sid", ""),
    ("Sinhala; Sinhalese", "si", "sin", ""),
    ("Siouan languages", "", "sio", ""),
    ("Sino-Tibetan languages", "", "sit", ""),
    ("Slavic languages", "", "sla", ""),
    ("Slovak", "sk", "slo", "slk"),
    ("Slovenian", "sl", "slv", ""),
    ("Southern Sami", "", "sma", ""),
    ("Northern Sami", "se", "sme", ""),
    ("Sami languages", "", "smi", ""),
    ("Lule Sami", "", "smj", ""),
    ("Inari Sami", "", "smn", ""),
    ("Samoan", "sm", "smo", ""),
    ("Skolt Sami", "", "sms", ""),
    ("Shona", "sn", "sna", ""),
    ("Sindhi", "sd", "snd", ""),
    ("Soninke", "", "snk", ""),
    ("Sogdian", "", "sog", ""),
    ("Somali", "so", "som", ""),
    ("Songhai languages", "", "son", ""),
    ("Sotho, Southern", "st", "sot", ""),
    ("Spanish; Castilian", "es", "spa", ""),
    ("Sardinian", "sc", "srd", ""),
    ("Sranan Tongo", "", "srn", ""),
    ("Serbian", "sr", "srp", ""),
    ("Serer", "", "srr", ""),
    ("Nilo-Saharan languages", "", "ssa", ""),
    ("Swati", "ss", "ssw", ""),
    ("Sukuma", "", "suk", ""),
    ("Sundanese", "su", "sun", ""),
    ("Susu", "", "sus", ""),
    ("Sumerian", "", "sux", ""),
    ("Swahili", "sw", "swa", ""),
    ("Swedish", "sv", "swe", ""),
    ("Classical Syriac", "", "syc", ""),
    ("Syriac", "", "syr", ""),
    ("Tahitian", "ty", "tah", ""),
    ("Tai languages", "", "tai", ""),
    ("Tamil", "ta", "tam", ""),
    ("Tatar", "tt", "tat", ""),
    ("Telugu", "te", "tel", ""),
    ("Timne", "", "tem", ""),
    ("Tereno", "", "ter", ""),
    ("Tetum", "", "tet", ""),
    ("Tajik", "tg", "tgk", ""),
    ("Tagalog", "tl", "tgl", ""),
    ("Thai", "th", "tha", ""),
    ("Tibetan", "bo", "tib", "bod"),
    ("Tigre", "", "tig", ""),
    ("Tigrinya", "ti", "tir", ""),
    ("Tiv", "", "tiv", ""),
    ("Tokelau", "", "tkl", ""),
    ("Klingon; tlhIngan-Hol", "", "tlh", ""),
    ("Tlingit", "", "tli", ""),
    ("Tamashek", "", "tmh", ""),
    ("Tonga (Nyasa)", "", "tog", ""),
    ("Tonga (Tonga Islands)", "to", "ton", ""),
    ("Tok Pisin", "", "tpi", ""),
    ("Tsimshian", "", "tsi", ""),
    ("Tswana", "tn", "tsn", ""),
    ("Tsonga", "ts", "tso", ""),
    ("Turkmen", "tk", "tuk", ""),
    ("Tumbuka", "", "tum", ""),
    ("Tupi languages", "", "tup", ""),
    ("Turkish", "tr", "tur", ""),
    ("Altaic languages", "", "tut", ""),
    ("Tuvalu", "", "tvl", ""),
    ("Twi", "tw", "twi", ""),
    ("Tuvinian", "", "tyv", ""),
    ("Udmurt", "", "udm", ""),
    ("Ugaritic", "", "uga", ""),
    ("Uighur; Uyghur", "ug", "uig", ""),
    ("Ukrainian", "uk", "ukr", ""),
    ("Umbundu", "", "umb", ""),
    ("Undetermined", "", "und", ""),
    ("Urdu", "ur", "urd", ""),
    ("Uzbek", "uz", "uzb", ""),
    ("Vai", "", "vai", ""),
    ("Venda", "ve", "ven", ""),
    ("Vietnamese", "vi", "vie", ""),
    ("Volap\xfck", "vo", "vol", ""),
    ("Votic", "", "vot", ""),
    ("Wakashan languages", "", "wak", ""),
    ("Wolaitta; Wolaytta", "", "wal", ""),
    ("Waray", "", "war", ""),
    ("Washo", "", "was", ""),
    ("Welsh", "cy", "wel", "cym"),
    ("Sorbian languages", "", "wen", ""),
    ("Walloon", "wa", "wln", ""),
    ("Wolof", "wo", "wol", ""),
    ("Kalmyk; Oirat", "", "xal", ""),
    ("Xhosa", "xh", "xho", ""),
    ("Yao", "", "yao", ""),
    ("Yapese", "", "yap", ""),
    ("Yiddish", "yi", "yid", ""),
    ("Yoruba", "yo", "yor", ""),
    ("Yupik languages", "", "ypk", ""),
    ("Zapotec", "", "zap", ""),
    ("Blissymbols; Blissymbolics; Bliss", "", "zbl", ""),
    ("Zenaga", "", "zen", ""),
    ("Standard Moroccan Tamazight", "", "zgh", ""),
    ("Zhuang; Chuang", "za", "zha", ""),
    ("Zande languages", "", "znd", ""),
    ("Zulu", "zu", "zul", ""),
    ("Zuni", "", "zun", ""),
    ("No linguistic content; Not applicable", "", "zxx", ""),
    ("Zaza; Dimili; Dimli; Kirdki; Kirmanjki; Zazaki", "", "zza", ""),
]

_ISO_639_1 = {}
_ISO_639_2 = {}
_LOWER = {}


def _fill_mappings():
    for entry in _ISO_639:
        name, _1, _2B, _2T = entry
        _LOWER[name.lower()] = name
        if _1:
            _ISO_639_1[_1] = entry
        _ISO_639_2[_2B] = entry
        if _2T:
            _ISO_639_2[_2T] = entry


_fill_mappings()


ISO_639_2 = list(_ISO_639_2.keys())


def _gettext(name, cache=None):
    cache = cache or []
    if not cache:
        t = register_translation("iso_639")
        cache.append(lambda s: t.wrap_text(t.gettext(s)))
    return cache[0](name)


def get_name(iso_code):
    """Returns the English name for the iso_639_1/2 code or an empty string
    if the code is not known.
    """

    if iso_code in _ISO_639_1:
        return _ISO_639_1[iso_code][0]
    if iso_code in _ISO_639_2:
        return _ISO_639_2[iso_code][0]

    return ""


def translate(text):
    """Given an iso code or a name tries to return a translated version
    for the current locale.

    If no translation is found returns an empty string.
    """

    # try to convert iso codes to English names and then translate
    name = get_name(text)
    if name:
        return _gettext(name)
    # not an iso code, try to match with the English name.
    # If all fails just return the original input
    lower = text.lower()
    if lower in _LOWER:
        return _gettext(_LOWER[lower])
    return ""


def _print_iso_639():
    """Prints the _ISO_639 list from above based on iso_639.xml.

    Maybe depend on the xml at runtime?
    """

    import os
    import pprint
    from xml.etree import cElementTree as ET  # noqa

    from .path import xdg_get_system_data_dirs

    entries = []
    for data_dir in xdg_get_system_data_dirs():
        iso_path = os.path.join(data_dir, "xml", "iso-codes", "iso_639.xml")
        if os.path.exists(iso_path):
            tree = ET.ElementTree(file=iso_path)
            for elm in tree.iterfind("iso_639_entry"):
                iso_639_1_code = elm.attrib.get("iso_639_1_code", "")
                iso_639_2B_code = elm.attrib["iso_639_2B_code"]
                iso_639_2T_code = elm.attrib["iso_639_2T_code"]
                if iso_639_2B_code == iso_639_2T_code:
                    iso_639_2T_code = ""
                name = elm.attrib["name"]
                entries.append((name, iso_639_1_code, iso_639_2B_code, iso_639_2T_code))
            break
    else:
        raise Exception("iso_639.xml not found")

    pprint.pprint(entries)



# Copyright 2014,2016 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import json
import os

from senf import fsn2text, uri2fsn

from quodlibet import C_, _
from quodlibet.util.dprint import print_, print_e
from quodlibet.remote import Remote, RemoteError


def exit_(status=None, notify_startup=False):
    """Call this to abort the startup before any mainloop starts.

    notify_startup needs to be true if QL could potentially have been
    called from the desktop file.
    """

    if notify_startup:
        import gi

        gi.require_version("Gdk", "3.0")
        from gi.repository import Gdk

        Gdk.notify_startup_complete()
    raise SystemExit(status)


def is_running():
    """If maybe is another instance running"""

    return Remote.remote_exists()


def control(command, arg=None, ignore_error=False):
    """Sends command to the existing instance if possible and exits.

    Will print any response it gets to stdout.

    Does not return except if ignore_error is True and sending
    the command failed.
    """

    if not is_running():
        if ignore_error:
            return
        exit_(
            _("Quod Libet is not running (add '--run' to start it)"),
            notify_startup=True,
        )
        return

    message = command
    if arg is not None:
        message += " " + arg

    try:
        response = Remote.send_message(message)
    except RemoteError as e:
        if ignore_error:
            return
        exit_(str(e), notify_startup=True)
    else:
        if response is not None:
            print_(response, end="", flush=True)
        exit_(notify_startup=True)


def process_arguments(argv):
    from quodlibet.util.path import uri_is_valid
    from quodlibet import util
    from quodlibet import const

    actions = []
    controls = [
        "next",
        "previous",
        "play",
        "pause",
        "play-pause",
        "stop",
        "hide-window",
        "show-window",
        "toggle-window",
        "focus",
        "quit",
        "unfilter",
        "refresh",
        "force-previous",
    ]
    controls_opt = [
        "seek",
        "repeat",
        "query",
        "volume",
        "filter",
        "rating",
        "set-browser",
        "open-browser",
        "shuffle",
        "queue",
        "stop-after",
        "random",
        "repeat-type",
        "shuffle-type",
        "add-location",
    ]

    options = util.OptionParser(
        "Quod Libet", const.VERSION, _("a music library and player"), _("[option]")
    )

    options.add("print-playing", help=_("Print the playing song and exit"))
    options.add("start-playing", help=_("Begin playing immediately"))
    options.add("start-hidden", help=_("Don't show any windows on start"))

    for opt, help in [
        ("next", _("Jump to next song")),
        ("previous", _("Jump to previous song or restart if near the beginning")),
        ("force-previous", _("Jump to previous song")),
        ("play", _("Start playback")),
        ("pause", _("Pause playback")),
        ("play-pause", _("Toggle play/pause mode")),
        ("stop", _("Stop playback")),
        ("volume-up", _("Turn up volume")),
        ("volume-down", _("Turn down volume")),
        ("rating-up", _("Increase rating of playing song by one star")),
        ("rating-down", _("Decrease rating of playing song by one star")),
        ("status", _("Print player status")),
        ("hide-window", _("Hide main window")),
        ("show-window", _("Show main window")),
        ("toggle-window", _("Toggle main window visibility")),
        ("focus", _("Focus the running player")),
        ("unfilter", _("Remove active browser filters")),
        ("refresh", _("Refresh and rescan library")),
        ("list-browsers", _("List available browsers")),
        ("print-playlist", _("Print the current playlist")),
        ("print-queue", _("Print the contents of the queue")),
        ("print-query-text", _("Print the active text query")),
        ("no-plugins", _("Start without plugins")),
        ("run", _("Start Quod Libet if it isn't running")),
        ("quit", _("Exit Quod Libet")),
    ]:
        options.add(opt, help=help)

    for opt, help, arg in [
        ("seek", _("Seek within the playing song"), _("[+|-][HH:]MM:SS")),
        ("shuffle", _("Set or toggle shuffle mode"), "0|1|t"),
        ("shuffle-type", _("Set shuffle mode type"), "random|weighted|off"),
        ("repeat", _("Turn repeat off, on, or toggle it"), "0|1|t"),
        ("repeat-type", _("Set repeat mode type"), "current|all|one|off"),
        ("volume", _("Set the volume"), "[+|-]0..100"),
        ("query", _("Search your audio library"), _("query")),
        ("play-file", _("Play a file"), C_("command", "filename")),
        ("rating", _("Set rating of playing song"), "[+|-]0.0..1.0"),
        ("set-browser", _("Set the current browser"), "BrowserName"),
        ("stop-after", _("Stop after the playing song"), "0|1|t"),
        ("open-browser", _("Open a new browser"), "BrowserName"),
        ("queue", _("Show or hide the queue"), "on|off|t"),
        ("random", _("Filter on a random value"), C_("command", "tag")),
        ("filter", _("Filter on a tag value"), _("tag=value")),
        (
            "enqueue",
            _("Enqueue a file or query"),
            "{}|{}".format(C_("command", "filename"), _("query")),
        ),
        (
            "enqueue-files",
            _("Enqueue comma-separated files"),
            "{}[,{}..]".format(_("filename"), _("filename")),
        ),
        ("print-query", _("Print filenames of results of query to stdout"), _("query")),
        (
            "unqueue",
            _("Unqueue a file or query"),
            "{}|{}".format(C_("command", "filename"), _("query")),
        ),
        ("add-location", _("Add a file or directory to the library"), _("location")),
        ("with-pattern", _("Set template for --print-* commands"), _("pattern")),
    ]:
        options.add(opt, help=help, arg=arg)

    options.add("sm-config-prefix", arg="dummy")
    options.add("sm-client-id", arg="prefix")
    options.add("screen", arg="dummy")

    def is_vol(str):
        if len(str) == 1 and str[0] in "+-":
            return True
        return is_float(str)

    def is_rate(str):
        if len(str) == 1 and str[0] in "+-":
            return True
        return is_float(str)

    def is_time(str):
        if str[0] not in "+-0123456789":
            return False
        if str[0] in "+-":
            str = str[1:]
        parts = str.split(":")
        if len(parts) > 3:
            return False
        return False not in [p.isdigit() for p in parts]

    def is_float(str):
        try:
            float(str)
        except ValueError:
            return False
        else:
            return True

    validators = {
        "shuffle": ["0", "1", "t", "on", "off", "toggle"].__contains__,
        "shuffle-type": ["random", "weighted", "off", "0"].__contains__,
        "repeat": ["0", "1", "t", "on", "off", "toggle"].__contains__,
        "repeat-type": ["current", "all", "one", "off", "0"].__contains__,
        "volume": is_vol,
        "seek": is_time,
        "rating": is_rate,
        "stop-after": ["0", "1", "t"].__contains__,
    }

    cmds_todo = []

    def queue(*args):
        cmds_todo.append(args)

    # XXX: to make startup work in case the desktop file isn't passed
    # a file path/uri
    if argv[-1] == "--play-file":
        argv = argv[:-1]

    opts, args = options.parse(argv[1:])

    for command, arg in opts.items():
        if command in controls:
            queue(command)
        elif command in controls_opt:
            if command in validators and not validators[command](arg):
                print_e(_("Invalid argument for '%s'.") % command)
                print_e(_("Try %s --help.") % fsn2text(argv[0]))
                exit_(True, notify_startup=True)
            else:
                queue(command, arg)
        elif command == "status":
            queue("status")
        elif command == "print-playlist":
            queue("dump-playlist", opts.get("with-pattern"))
        elif command == "print-queue":
            queue("dump-queue", opts.get("with-pattern"))
        elif command == "list-browsers":
            queue("dump-browsers")
        elif command == "volume-up":
            queue("volume +")
        elif command == "volume-down":
            queue("volume -")
        elif command == "rating-up":
            queue("rating +")
        elif command == "rating-down":
            queue("rating -")
        elif command == "enqueue" or command == "unqueue":
            try:
                filename = uri2fsn(arg)
            except ValueError:
                filename = arg
            queue(command, filename)
        elif command == "enqueue-files":
            queue(command, arg)
        elif command == "play-file":
            if uri_is_valid(arg) and arg.startswith("quodlibet://"):
                # TODO: allow handling of URIs without --play-file
                queue("uri-received", arg)
            else:
                try:
                    filename = uri2fsn(arg)
                except ValueError:
                    filename = arg
                filename = os.path.abspath(os.path.expanduser(arg))
                queue("play-file", filename)
        elif command == "add-location":
            try:
                path = uri2fsn(arg)
            except ValueError:
                path = arg
            path = os.path.abspath(os.path.expanduser(arg))
            queue("add-location", path)
        elif command == "print-playing":
            try:
                queue("print-playing", args[0])
            except IndexError:
                queue("print-playing", opts.get("with-pattern"))
        elif command == "print-query":
            pattern = opts.get("with-pattern")
            queue(command, json.dumps({"query": arg, "pattern": pattern}))
        elif command == "print-query-text":
            queue(command)
        elif command == "start-playing":
            actions.append(command)
        elif command == "start-hidden":
            actions.append(command)
        elif command == "no-plugins":
            actions.append(command)
        elif command == "run":
            actions.append(command)

    if cmds_todo:
        for cmd in cmds_todo:
            control(*cmd, ignore_error="run" in actions)
    else:
        # this will exit if it succeeds
        control("focus", ignore_error=True)

    return actions, cmds_todo



# Copyright 2011 Christoph Reiter <reiter.christoph@gmail.com>
# Copyright 2020 Antigone <mail@antigone.xyz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import sys

if os.name == "nt" or sys.platform == "darwin":
    from quodlibet.plugins import PluginNotSupportedError

    raise PluginNotSupportedError

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk

from quodlibet import _
from quodlibet import app
from quodlibet import config
from quodlibet.qltk import Icons
from quodlibet.plugins.events import EventPlugin


def get_toplevel_xid():
    if app.window.get_window():
        try:
            return app.window.get_window().get_xid()
        except AttributeError:  # non x11
            pass
    return 0


class InhibitFlags:
    LOGOUT = 1
    USERSWITCH = 1 << 1
    SUSPEND = 1 << 2
    IDLE = 1 << 3


class InhibitStrings:
    SUSPEND = "inhibit_suspend"
    IDLE = "inhibit_idle"


class SessionInhibit(EventPlugin):
    PLUGIN_ID = "screensaver_inhibit"
    PLUGIN_NAME = _("Inhibit Screensaver/Suspend")
    PLUGIN_DESC = _(
        "On a GNOME desktop, when a song is playing, prevents"
        " either the screensaver from activating, or prevents the"
        " computer from suspending."
    )
    PLUGIN_ICON = Icons.PREFERENCES_DESKTOP_SCREENSAVER

    CONFIG_MODE = PLUGIN_ID + "_mode"

    DBUS_NAME = "org.gnome.SessionManager"
    DBUS_INTERFACE = "org.gnome.SessionManager"
    DBUS_PATH = "/org/gnome/SessionManager"

    APPLICATION_ID = "quodlibet"
    INHIBIT_REASON = _("Music is playing")

    __cookie = None

    def __get_dbus_proxy(self):
        bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
        return Gio.DBusProxy.new_sync(
            bus,
            Gio.DBusProxyFlags.NONE,
            None,
            self.DBUS_NAME,
            self.DBUS_PATH,
            self.DBUS_INTERFACE,
            None,
        )

    def enabled(self):
        if not app.player.paused:
            self.plugin_on_unpaused()

    def disabled(self):
        if not app.player.paused:
            self.plugin_on_paused()

    def plugin_on_unpaused(self):
        xid = get_toplevel_xid()
        mode = config.get("plugins", self.CONFIG_MODE, InhibitStrings.IDLE)
        flags = (
            InhibitFlags.SUSPEND
            if mode == InhibitStrings.SUSPEND
            else InhibitFlags.IDLE
        )

        try:
            dbus_proxy = self.__get_dbus_proxy()
            self.__cookie = dbus_proxy.Inhibit(
                "(susu)", self.APPLICATION_ID, xid, self.INHIBIT_REASON, flags
            )
        except GLib.Error:
            pass

    def plugin_on_paused(self):
        if self.__cookie is None:
            return

        try:
            dbus_proxy = self.__get_dbus_proxy()
            dbus_proxy.Uninhibit("(u)", self.__cookie)
            self.__cookie = None
        except GLib.Error:
            pass

    def PluginPreferences(self, parent):
        def changed(combo):
            index = combo.get_active()
            mode = InhibitStrings.SUSPEND if index == 1 else InhibitStrings.IDLE
            config.set("plugins", self.CONFIG_MODE, mode)
            if not app.player.paused:
                self.plugin_on_paused()
                self.plugin_on_unpaused()

        mode = config.get("plugins", self.CONFIG_MODE, InhibitStrings.IDLE)

        hb = Gtk.HBox(spacing=6)
        hb.set_border_width(6)
        # Translators: Inhibiting Mode
        hb.pack_start(Gtk.Label(label=_("Mode:")), False, True, 0)
        combo = Gtk.ComboBoxText()
        combo.append_text(_("Inhibit Screensaver"))
        combo.append_text(_("Inhibit Suspend"))
        combo.set_active(1 if mode == InhibitStrings.SUSPEND else 0)
        combo.connect("changed", changed)
        hb.pack_start(combo, True, True, 0)
        return hb



# Copyright 2013 Simonas Kazlauskas
#           2022 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from os import path, makedirs
from hashlib import sha1

from gi.repository import GObject

from quodlibet import get_cache_dir
from quodlibet.qltk import Icons
from quodlibet.util.path import escape_filename
from quodlibet.util import print_w


class CoverSourcePlugin(GObject.Object):
    """
    Plugins that given a song should provide a cover art.

    The plugin should override following methods and properties:

        @staticmethod priority()
        @property cover_path(self)
        fetch_cover(self)

    Refer to default function implementation's documentation in order to
    understand their role.
    """

    __gsignals__ = {
        "fetch-success": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "fetch-failure": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "search-complete": (GObject.SignalFlags.RUN_LAST, None, (object,)),
    }
    PLUGIN_ICON = Icons.EMBLEM_DOWNLOADS

    @property
    def name(self):
        """Human name of the source"""
        return type(self).__name__

    def __str__(self):
        return f"<{self.name} for {self.group_by(self.song)!r}>"

    embedded = False
    """Whether the source is an embedded one"""

    def __init__(self, song, cancellable=None):
        self.song = song
        self.cancellable = cancellable
        super().__init__()

    @classmethod
    def group_by(cls, song):
        """Returns a hashable for a song, for grouping songs in groups where
        only one song per group needs to be searched.

        Grouping might reduce the chance of finding covers in exchange
        for performance.

        This default implementation gives one group for all songs.
        """

        return

    @staticmethod
    def priority():
        """
        Should return float in range [0.0, 1.0] suggesting priority of the
        cover source. Whether value returned by this method is respected or
        not is not guaranteed.

        As a rule of thumb, source's reliability and quality should be
        compared with other sources and given score between two sources that
        come close in quality and reliability.

        There's a table of value ranges sources should respect:

        * (0.7, 1.0] - local covers;
        * (0.4, 0.7] - accurate (> 99%) source of high quality (>= 200x200)
                       covers;
        * (0.2, 0.4] - accurate (> 99%) source of low quality (< 200x200)
                       covers;
        * (0.0, 0.2] - not very accurate (<= 99%) source of covers, even if
                       they're high quality;
        *  0.0       - reserved for the fallback cover source.
        """
        return 0.0

    @property
    def cover_directory(self):
        return cover_dir

    @property
    def cover_filename(self):
        """
        Return the filename of the cover which hopefully should not change
        between songs in the same album and still be unique enough to
        uniquely identify most (or even better – all) of the albums.

        The string returned must not contain any characters illegal in
        most common filesystems. These include /, ?, <, >, \\, :, *, |, ” and
        ^. Staying in the bounds of ASCII is highly encouraged.

        Perchance the song lacks data to generate the filename of cover for
        this provider, None shall be returned.
        """
        key = sha1()
        # Should be fine as long as the same interpreter is used.
        data = repr(self.song.album_key)
        if not isinstance(data, bytes):
            data = data.encode("utf-8")
        key.update(data)
        return escape_filename(key.hexdigest())

    @property
    def cover_path(self):
        """
        Should return the path where cover is expected to be cached. The
        location should be based in common cache location available in variable
        `cover_dir` of this module.

        It doesn't necessarily mean the cover is actually at the returned
        location neither that it will be stored there at any later time.
        """
        return path.join(self.cover_directory, self.cover_filename)

    @property
    def cover(self):
        """
        Method to get cover file from cover provider for a specific song.

        Should always return a file-like object opened as read-only if any
        and None otherwise.
        """
        cp = self.cover_path
        try:
            return open(cp, "rb") if cp and path.isfile(cp) else None
        except OSError:
            print_w(f'Failed reading album art "{path}"')

    def search(self):
        """
        Start searching for cover art from a source.

        After search is completed the `search-complete` event must be emitted
        regardless of search outcome with a list of dictionaries containing
        `album`, `artist` and `cover` keys as an argument. If search was
        unsuccessful, empty list should be returned.

        By convention better quality and more accurate covers are expected to
        appear first in the list.
        """
        self.emit("search-complete", [])

    def fetch_cover(self):
        """
        Method to ask source fetch the cover from its source into location at
        `self.cover_path`.

        If this method succeeds in putting the image from its source into
        `self.cover_path`, `fetch-success` signal shall be emitted and
        `fetch-failure` otherwise.

        Return value of this function doesn't have any meaning whatsoever.
        """
        self.fail("This source is incapable of fetching covers")

    def fail(self, message):
        """
        Shorthand method for emitting `fetch-failure` signals.

        Most common use pattern would be:
            return self.fail("Failure message")
        """
        self.emit("fetch-failure", message)


cover_dir = path.join(get_cache_dir(), "covers")

try:
    makedirs(cover_dir)
except OSError:
    pass



# Copyright 2011,2013,2014 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import time
import threading

from quodlibet import config


def get_api_key():
    """The user API key for submissions"""

    return config.get("plugins", "fingerprint_acoustid_api_key", "")


def get_write_mb_tags():
    return config.getboolean("plugins", "fingerprint_write_mb_tags", False)


def get_group_by_dir():
    return config.getboolean("plugins", "fingerprint_group_by_dir", True)


class GateKeeper:
    def __init__(self, requests_per_sec):
        self._period = 1 / float(requests_per_sec)
        self._last = 0
        self._lock = threading.Lock()

    def wait(self):
        """Block until a new request can be made."""

        while 1:
            with self._lock:
                current = time.time()
                if abs(current - self._last) >= self._period:
                    self._last = current
                    break
                time.sleep(self._period / 10)



# Copyright 2006 Joe Wreschnig
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Various library classes.

Libraries are (roughly) dicts of items, and in the case of Quod Libet,
these items are songs. Libraries are in charge of alerting the rest of
the program when songs have been added, changed, or removed. They can
also be queried in various ways.
"""

import time

from quodlibet import print_d, print_w, config

from quodlibet.library.song import SongLibrary, SongFileLibrary
from quodlibet.library.librarians import SongLibrarian
from quodlibet.util.library import get_scan_dirs
from quodlibet.util.path import mtime


def init(cache_fn=None):
    """Set up the library and return the main one.

    Return a main library, and set a librarian for
    all future SongLibraries.
    """

    SongFileLibrary.librarian = SongLibrary.librarian = SongLibrarian()
    watch = config.getboolean("library", "watch")
    library = SongFileLibrary("main", watch_dirs=get_scan_dirs() if watch else [])
    if cache_fn:
        library.load(cache_fn)
    return library


def save(save_period=None):
    """Save all registered libraries that have a filename and are marked dirty.

    If `save_period` (seconds) is given the library will only be saved if
    it hasn't been in the last `save_period` seconds.
    """

    print_d("Saving all libraries...")

    librarian = SongFileLibrary.librarian
    for lib in librarian.libraries.values():
        filename = lib.filename
        if not filename or not lib.dirty:
            continue

        if not save_period or abs(time.time() - mtime(filename)) > save_period:
            lib.save()


def destroy() -> None:
    """Destroy all registered libraries"""

    print_d("Destroying all libraries...")

    librarian = SongFileLibrary.librarian
    if librarian:
        for lib in list(librarian.libraries.values()):
            try:
                lib.destroy()
            except Exception as e:
                print_w(f"Couldn't destroy {lib} ({e!r})")
        librarian.destroy()



# Copyright 2014 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
import struct
import os

from senf import path2fsn, fsn2text
from ._audio import AudioFile, translate_errors
from quodlibet.util.dprint import print_d
from quodlibet.util import list_unique

# VGM and GD3 SPECs:
# http://www.smspower.org/uploads/Music/vgmspec170.txt
# http://www.smspower.org/uploads/Music/gd3spec100.txt
GD3_TAG_PTR_POS = 0x14
GD3_TAG_PTR_SIZE = 4

GD3_ENGLISH_TITLE = 0
GD3_JAPANESE_TITLE = 1
GD3_ENGLISH_GAME = 2
GD3_JAPANESE_GAME = 3
GD3_ENGLISH_SYSTEM = 4
GD3_JAPANESE_SYSTEM = 5
GD3_ENGLISH_ARTIST = 6
GD3_JAPANESE_ARTIST = 7
GD3_DATE = 8
GD3_DUMPER = 9
GD3_COMMENT = 10


class VgmFile(AudioFile):
    format = "VGM"
    mimes: list[str] = []

    def __init__(self, filename):
        with translate_errors():
            with open(filename, "rb") as h:
                header = h.read(64)
                if len(header) != 64 or header[:4] != b"Vgm ":
                    # filename useful to show (helps w/ faulty VGM files.)
                    raise Exception(filename + " not a VGM file")

                def samples_to_sec(s):
                    return s / 44100.0

                samples = struct.unpack("<i", header[24:28])[0]
                loop_offset = struct.unpack("<i", header[28:32])[0]
                loop_samples = struct.unpack("<i", header[32:36])[0]

                # this should match libgme
                length = samples_to_sec(samples)
                if length <= 0:
                    length = 150
                elif loop_offset:
                    # intro + 2 loops
                    length += samples_to_sec(loop_samples)

                self["~#length"] = length

                gd3_position = struct.unpack(
                    "<i", header[GD3_TAG_PTR_POS : GD3_TAG_PTR_POS + GD3_TAG_PTR_SIZE]
                )[0]
                h.seek(GD3_TAG_PTR_POS + gd3_position)
                self.update(parse_gd3(h.read()))

        self.setdefault("title", fsn2text(path2fsn(os.path.basename(filename)[:-4])))
        self.sanitize(filename)

    def write(self):
        pass

    def can_change(self, k=None):
        if k is None:
            return ["title"]
        return k == "title"


def parse_gd3(data):
    tags = {}
    if data[0:4] != b"Gd3 ":
        print_d("Invalid Gd3, Missing Header...")
        return tags

    gd3_length = struct.unpack("<i", data[8:12])[0]
    # Length of gd3 footer. This means we can actually add more tags to the end
    # with APETAG at some point. (Or at least consider...)

    entries = data[12 : 12 + gd3_length].decode("utf-16-le").split("\0")

    if len(entries) > GD3_JAPANESE_TITLE:
        titles = gd3_filter_entries(
            [entries[GD3_ENGLISH_TITLE], entries[GD3_JAPANESE_TITLE]]
        )
        if len(titles) > 0:
            tags["title"] = "\n".join(titles)

    if len(entries) > GD3_JAPANESE_ARTIST:
        artists = gd3_filter_entries(
            [entries[GD3_ENGLISH_ARTIST], entries[GD3_JAPANESE_ARTIST]]
        )
        if len(artists) > 0:
            tags["artist"] = "\n".join(artists)

    if len(entries) > GD3_JAPANESE_SYSTEM:
        consoles = gd3_filter_entries(
            [entries[GD3_ENGLISH_SYSTEM], entries[GD3_JAPANESE_SYSTEM]]
        )
        if len(consoles) > 0:
            tags["console"] = "\n".join(consoles)

    if len(entries) > GD3_JAPANESE_GAME:
        games = gd3_filter_entries(
            [entries[GD3_ENGLISH_GAME], entries[GD3_JAPANESE_GAME]]
        )
        if len(games) > 0:
            tags["album"] = "\n".join(games)

    if len(entries) > GD3_DATE and entries[GD3_DATE] != "":
        tags["date"] = entries[GD3_DATE]

    if len(entries) > GD3_DUMPER and entries[GD3_DUMPER] != "":
        tags["dumper"] = entries[GD3_DUMPER]

    if len(entries) > GD3_COMMENT:
        tags["comment"] = entries[GD3_COMMENT]

    return tags


def gd3_filter_entries(entries):
    # First, filter out empty strings...
    filtered = list(filter(None, entries))
    if len(filtered) == 0:
        return filtered

    # Then, filter out any duplicate strings...
    filtered = list_unique(filtered)
    return sorted(filtered)


loader = VgmFile
types = [VgmFile]
extensions = [".vgm"]



# Copyright 2009 Steven Robertson
#        2016-20 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from quodlibet import _
from quodlibet import app
from quodlibet import config
from quodlibet import qltk
from quodlibet.plugins.playorder import ShufflePlugin
from quodlibet.qltk.ccb import ConfigCheckButton
from quodlibet.qltk import Icons
from quodlibet.qltk.playorder import OrderInOrder

from gi.repository import Gtk


class QueueOrder(ShufflePlugin, OrderInOrder):
    PLUGIN_ID = "queue"
    PLUGIN_NAME = _("Queue Only")
    PLUGIN_ICON = Icons.VIEW_LIST
    PLUGIN_DESC = _(
        "Limits playing of songs to the queue.\n\n"
        "Select this play order in the main window, "
        "then double-clicking any song will enqueue it "
        "instead of playing."
    )
    display_name = _("Queue only")
    accelerated_name = _("_Queue only")

    def PluginPreferences(self):
        box = Gtk.HBox()
        ccb = ConfigCheckButton(
            _("Automatically start playing double-clicked songs"),
            "plugins",
            "queue_only_autoplay",
        )
        autoplay = config.getboolean("plugins", "queue_only_autoplay", False)
        ccb.set_active(autoplay)
        box.pack_start(qltk.Frame(_("Preferences"), child=ccb), True, True, 0)
        return box

    def next(self, playlist, iter):
        return None

    def set_explicit(self, playlist, iter):
        if iter is None:
            return
        song = playlist[iter][0]
        if song is None:
            return

        # store queue size for jumping to newly added items
        queue = app.window.playlist.q
        queue_size = len(queue)

        app.window.playlist.enqueue([playlist[iter][0]])

        # if setting enabled, start playing the first added song
        autoplay = config.getboolean("plugins", "queue_only_autoplay", False)
        if autoplay and len(queue) > queue_size:
            # queue_size is 1 greater than previous last index
            new_song_iter = queue.iter_nth_child(None, queue_size)
            app.window.playlist._player.go_to(new_song_iter, True, queue)
            app.window.playlist._player.play()



# Copyright 2016-23 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import time
import operator
from collections import defaultdict
from datetime import datetime

from quodlibet import print_d
from quodlibet.formats import AudioFile
from quodlibet.query import Query, QueryType
from quodlibet.query._match import (
    Tag,
    Inter,
    Union,
    Numcmp,
    NumexprTag,
    Numexpr,
    True_,
    False_,
)

INVERSE_OPS = {
    operator.le: operator.gt,
    operator.gt: operator.le,
    operator.lt: operator.ge,
    operator.ge: operator.lt,
}

_DUMMY_AF = AudioFile()
_CLOCK = time.time


def convert_time(t):
    return datetime.strftime(datetime.fromtimestamp(int(t)), "%Y-%m-%d %H:%S")


_QL_TO_SC = {
    "genre": ("genres", None),
    "length": ("duration", lambda x: int((x or 0) * 1000)),
    "date": ("created_at", convert_time),
    "tags": ("tags", None),
    "bpm": ("bpm", None),
    "artist": ("q", None),
    "title": ("q", None),
    "comments": ("q", None),
    "soundcloud_user_id": ("user_id", None),
}
""" Convert QL to Soundcloud tags with optional value mapper"""

SUPPORTED = set(_QL_TO_SC.keys()) | {"rating"}


class SoundcloudQuery(Query):
    def __init__(self, string, star=None, clock=time.time):
        super().__init__(string, star)
        self._clock = clock
        try:
            self.terms = self._extract_terms(self._match)
        except self.Error as e:
            print_d(f"Couldn't use query: {e}")
            self.type = QueryType.INVALID
            self.terms = {}

    def _extract_terms(self, node):
        """Return a dict of sets keyed on API search term,
        with values for these that could be used to query the API
        and might return results useful for populating the songlist.

        Note this is not a *translation* of the query in any sense,
        and that (currently) the browser filters ingested API results
        so that the QL results are still valid based on
        the query given, even if some more could have been returned.

        ...so if in doubt, *less* restrictive is better here."""
        tuples = self._extract_terms_set(node)
        terms = defaultdict(set)
        for k, v in tuples:
            terms[k].add(v)
        return terms

    def _extract_terms_set(self, node, tag=None):
        def to_api(tag, raw_value):
            try:
                api_tag, converter = _QL_TO_SC[tag] if tag else ("q", None)
            except KeyError as e:
                if tag not in SUPPORTED:
                    raise self.Error(
                        f"Unsupported {tag!r} tag. Try: {', '.join(SUPPORTED)}"
                    ) from e
                return None, None
            else:
                value = str(converter(raw_value) if converter else raw_value)
                return api_tag, value

        def terms_from_re(pattern, t):
            """Best efforts to de-regex"""
            pat = pattern.lstrip("^").rstrip("$")
            api_tag, pat = to_api(t, pat)
            return {(api_tag, p) for p in pat.split("|")} if api_tag else set()

        if isinstance(node, Tag) and set(node._names) & SUPPORTED:
            if len(node._names) == 1:
                return self._extract_terms_set(node.res, tag=node._names[0])
            return self._extract_terms_set(node.res)
        if isinstance(node, Inter | Union):
            # Treat identically as the text-based query will perform
            # relevance ranking itself, meaning that any term is still useful
            terms = set()
            for n in node.res:
                terms |= self._extract_terms_set(n)
            return terms
        if isinstance(node, Numcmp):

            def from_relative(op, l, r):
                raw_value = r.evaluate(_DUMMY_AF, self._clock(), True)
                tag, value = to_api(l._tag, raw_value)
                if not value:
                    return set()
                if op == operator.eq:
                    return {(tag, value)}
                if op in (operator.le, operator.lt):
                    return {(tag + "[to]", value)}
                if op in (operator.ge, operator.gt):
                    return {(tag + "[from]", value)}
                raise self.Error(f"Unsupported operator: {op}")

            left = node._expr
            right = node._expr2
            if isinstance(left, NumexprTag) and isinstance(right, Numexpr):
                return from_relative(node._op, left, right)
            if isinstance(right, NumexprTag) and isinstance(left, Numexpr):
                # We can reduce the logic by flipping the expression
                return from_relative(INVERSE_OPS[node._op], right, left)
            raise self.Error(f"Unsupported numeric: {node}")
        if hasattr(node, "pattern"):
            return terms_from_re(node.pattern, tag)
        if isinstance(node, True_):
            return set()
        if isinstance(node, False_):
            raise self.Error("False can never be queried")
        raise self.Error(f"Unhandled node: {node!r}")



# Copyright 2004-2005 Joe Wreschnig, Michael Urman, Iñigo Serna
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import re
import os

from gi.repository import Gtk
from senf import fsn2text

import quodlibet

from quodlibet import _, ngettext
from quodlibet import config
from quodlibet import qltk
from quodlibet import util

from quodlibet.formats import AudioFileError
from quodlibet.plugins import PluginManager
from quodlibet.qltk._editutils import FilterPluginBox, FilterCheckButton
from quodlibet.qltk._editutils import EditingPluginHandler, OverwriteWarning
from quodlibet.qltk._editutils import WriteFailedError
from quodlibet.qltk.wlw import WritingWindow
from quodlibet.qltk.views import TreeViewColumn
from quodlibet.qltk.cbes import ComboBoxEntrySave
from quodlibet.qltk.models import ObjectStore
from quodlibet.qltk import Icons
from quodlibet.util.tagsfrompath import TagsFromPattern
from quodlibet.util.string.splitters import split_value
from quodlibet.util import connect_obj
from quodlibet.plugins.editing import TagsFromPathPlugin


TBP = os.path.join(quodlibet.get_user_dir(), "lists", "tagpatterns")
TBP_EXAMPLES = """\
<tracknumber>. <title>
<tracknumber> - <title>
<tracknumber> - <artist> - <title>
<artist> - <album>/<tracknumber>. <title>
<artist>/<album>/<tracknumber> - <title>"""


class UnderscoresToSpaces(FilterCheckButton):
    _label = _("Replace _underscores with spaces")
    _section = "tagsfrompath"
    _key = "underscores"
    _order = 1.0

    def filter(self, tag, value):
        return value.replace("_", " ")


class TitleCase(FilterCheckButton):
    _label = _("_Title-case tags")
    _section = "tagsfrompath"
    _key = "titlecase"
    _order = 1.1

    def filter(self, tag, value):
        return util.title(value)


class SplitTag(FilterCheckButton):
    _label = _("Split into multiple _values")
    _section = "tagsfrompath"
    _key = "split"
    _order = 1.2

    def filter(self, tag, value):
        spls = config.gettext("editing", "split_on")
        spls = spls.split()
        return "\n".join(split_value(value, spls))


class TagsFromPathPluginHandler(EditingPluginHandler):
    Kind = TagsFromPathPlugin


class ListEntry:
    def __init__(self, song):
        self.song = song
        self.matches = {}

    def get_match(self, key):
        return self.matches.get(key, "")

    def replace_match(self, key, value):
        self.matches[key] = value

    @property
    def name(self):
        return fsn2text(self.song("~basename"))


class TagsFromPath(Gtk.VBox):
    title = _("Tags From Path")
    FILTERS = [UnderscoresToSpaces, TitleCase, SplitTag]
    handler = TagsFromPathPluginHandler()

    @classmethod
    def init_plugins(cls):
        PluginManager.instance.register_handler(cls.handler)

    def __init__(self, parent, library):
        super().__init__(spacing=6)

        self.set_border_width(12)
        hbox = Gtk.HBox(spacing=6)
        cbes_defaults = TBP_EXAMPLES.split("\n")
        self.combo = ComboBoxEntrySave(
            TBP,
            cbes_defaults,
            title=_("Path Patterns"),
            edit_title=_("Edit saved patterns…"),
        )
        self.combo.show_all()
        hbox.pack_start(self.combo, True, True, 0)
        self.preview = qltk.Button(_("_Preview"), Icons.VIEW_REFRESH)
        self.preview.show()
        hbox.pack_start(self.preview, False, True, 0)
        self.pack_start(hbox, False, True, 0)
        self.combo.get_child().connect("changed", self._changed)

        model = ObjectStore()
        self.view = Gtk.TreeView(model=model)
        self.view.show()

        sw = Gtk.ScrolledWindow()
        sw.set_shadow_type(Gtk.ShadowType.IN)
        sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        sw.add(self.view)
        self.pack_start(sw, True, True, 0)

        vbox = Gtk.VBox()
        addreplace = Gtk.ComboBoxText()
        addreplace.append_text(_("Tags replace existing ones"))
        addreplace.append_text(_("Tags are added to existing ones"))
        addreplace.set_active(config.getboolean("tagsfrompath", "add"))
        addreplace.connect("changed", self.__add_changed)
        vbox.pack_start(addreplace, True, True, 0)
        addreplace.show()
        self.pack_start(vbox, False, True, 0)

        filter_box = FilterPluginBox(self.handler, self.FILTERS)
        filter_box.connect("preview", self.__filter_preview)
        filter_box.connect("changed", self.__filter_changed)
        self.filter_box = filter_box
        self.pack_start(filter_box, False, True, 0)

        # Save button
        self.save = qltk.Button(_("_Save"), Icons.DOCUMENT_SAVE)
        self.save.show()
        bbox = Gtk.HButtonBox()
        bbox.set_layout(Gtk.ButtonBoxStyle.END)
        bbox.pack_start(self.save, True, True, 0)
        self.pack_start(bbox, False, True, 0)

        connect_obj(self.preview, "clicked", self.__preview, None)
        connect_obj(parent, "changed", self.__class__.__preview, self)

        # Save changes
        connect_obj(self.save, "clicked", self.__save, addreplace, library)

        for child in self.get_children():
            child.show()

    def __filter_preview(self, *args):
        Gtk.Button.clicked(self.preview)

    def __filter_changed(self, *args):
        self._changed(self.combo.get_child())

    def _changed(self, entry):
        self.save.set_sensitive(False)
        self.preview.set_sensitive(bool(entry.get_text()))

    def __add_changed(self, combo):
        config.set("tagsfrompath", "add", str(bool(combo.get_active())))

    def __preview(self, songs):
        if songs is None:
            songs = [row[0].song for row in (self.view.get_model() or [])]

        if songs:
            pattern_text = self.combo.get_child().get_text()
        else:
            pattern_text = ""
        try:
            pattern = TagsFromPattern(pattern_text)
        except re.error:
            qltk.ErrorMessage(
                self,
                _("Invalid pattern"),
                _(
                    "The pattern\n\t%s\nis invalid. "
                    "Possibly it contains the same tag twice or "
                    "it has unbalanced brackets (&lt; / &gt;)."
                )
                % (util.bold(pattern_text)),
                escape_desc=False,
            ).run()
            return
        else:
            if pattern_text:
                self.combo.prepend_text(pattern_text)
                self.combo.write(TBP)

        invalid = []

        for header in pattern.headers:
            if not min([song.can_change(header) for song in songs]):
                invalid.append(header)
        total = len(invalid)
        if total and songs:
            title = ngettext("Invalid tag", "Invalid tags", total)
            msg = ngettext(
                "Invalid tag %s\n\nThe files currently "
                "selected do not support editing this tag.",
                "Invalid tags %s\n\nThe files currently "
                "selected do not support editing these tags.",
                total,
            )
            tags_str = util.bold(", ".join(invalid))
            qltk.ErrorMessage(self, title, msg % tags_str, escape_desc=False).run()
            pattern = TagsFromPattern("")

        self.view.set_model(None)
        model = ObjectStore()
        for col in self.view.get_columns():
            self.view.remove_column(col)

        render = Gtk.CellRendererText()
        col = TreeViewColumn(title=_("File"))
        col.pack_start(render, True)
        col.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)

        def cell_data_file(column, cell, model, iter_, data):
            entry = model.get_value(iter_)
            cell.set_property("text", entry.name)

        col.set_cell_data_func(render, cell_data_file)

        def cell_data_header(column, cell, model, iter_, header):
            entry = model.get_value(iter_)
            cell.set_property("text", entry.get_match(header))

        self.view.append_column(col)
        for _i, header in enumerate(pattern.headers):
            render = Gtk.CellRendererText()
            render.set_property("editable", True)
            render.connect("edited", self.__row_edited, model, header)
            escaped_title = header.replace("_", "__")
            col = Gtk.TreeViewColumn(escaped_title, render)
            col.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
            col.set_cell_data_func(render, cell_data_header, header)
            self.view.append_column(col)

        for song in songs:
            entry = ListEntry(song)
            match = pattern.match(song)
            for h in pattern.headers:
                text = match.get(h, "")
                for f in self.filter_box.filters:
                    if f.active:
                        text = f.filter(h, text)
                if not song.can_multiple_values(h):
                    text = ", ".join(text.split("\n"))
                entry.matches[h] = text
            model.append([entry])

        # save for last to potentially save time
        if songs:
            self.view.set_model(model)
        self.preview.set_sensitive(False)
        self.save.set_sensitive(len(pattern.headers) > 0)

    def __save(self, addreplace, library):
        pattern_text = self.combo.get_child().get_text()
        pattern = TagsFromPattern(pattern_text)
        model = self.view.get_model()
        add = bool(addreplace.get_active())
        win = WritingWindow(self, len(model))
        win.show()

        was_changed = set()

        all_done = False
        for entry in (model and model.values()) or []:
            song = entry.song
            changed = False
            if not song.valid():
                win.hide()
                dialog = OverwriteWarning(self, song)
                resp = dialog.run()
                win.show()
                if resp != OverwriteWarning.RESPONSE_SAVE:
                    break

            for _i, h in enumerate(pattern.headers):
                text = entry.get_match(h)
                if text:
                    can_multiple = song.can_multiple_values(h)
                    if not add or h not in song or not can_multiple:
                        song[h] = text
                        changed = True
                    else:
                        for val in text.split("\n"):
                            if val not in song.list(h):
                                song.add(h, val)
                                changed = True

            if changed:
                try:
                    song.write()
                except AudioFileError:
                    util.print_exc()
                    WriteFailedError(self, song).run()
                    library.reload(song, changed=was_changed)
                    break
                was_changed.add(song)

            if win.step():
                break
        else:
            all_done = True

        win.destroy()
        library.changed(was_changed)
        self.save.set_sensitive(not all_done)

    def __row_edited(self, renderer, path, new, model, header):
        entry = model[path][0]
        if entry.get_match(header) != new:
            entry.replace_match(header, new)
            self.preview.set_sensitive(True)
            self.save.set_sensitive(True)



# Copyright 2006 Joe Wreschnig
#           2014 Christoph Reiter
#           2016 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os

from gi.repository import Gio

from quodlibet import _
from quodlibet.qltk import Icons
from quodlibet import app
from quodlibet.plugins.events import EventPlugin


class AutoMasking(EventPlugin):
    PLUGIN_ID = "automask"
    PLUGIN_NAME = _("Automatic Masking")
    PLUGIN_DESC = _(
        "Automatically masks and unmasks drives when they are unmounted or mounted."
    )
    PLUGIN_ICON = Icons.DRIVE_REMOVABLE_MEDIA

    __sigs = None
    __monitor = None

    def enabled(self):
        if self.__monitor is None:
            self.__monitor = Gio.VolumeMonitor.get()
            self.__sigs = [
                self.__monitor.connect("mount-added", self.__mounted),
                self.__monitor.connect("mount-removed", self.__unmounted),
            ]
        else:
            for signal_id in self.__sigs:
                self.__monitor.handler_unblock(signal_id)

    def disabled(self):
        for signal_id in self.__sigs:
            self.__monitor.handler_unblock(signal_id)

    def __mounted(self, monitor, mount):
        path = mount.get_default_location().get_path()
        if path is not None:
            app.library.unmask(os.path.normpath(path))

    def __unmounted(self, monitor, mount):
        path = mount.get_default_location().get_path()
        if path is not None:
            app.library.mask(os.path.normpath(path))



# Copyright 2014 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from collections.abc import Callable
import re
import sre_parse
import sre_constants
import unicodedata

from quodlibet import print_d
from quodlibet.util import re_escape

from .db import get_replacement_mapping


def _fixup_literal(literal, in_seq, mapping):
    u = chr(literal)
    if u in mapping:
        u = u + "".join(mapping[u])
    need_seq = len(u) > 1
    u = re_escape(u)
    if need_seq and not in_seq:
        u = f"[{u}]"
    return u


def _fixup_literal_list(literals, mapping):
    u = "".join(map(chr, literals))

    # longest matches first, we will handle contained ones in the replacement
    # function
    reg = "({})".format(
        "|".join(map(re_escape, sorted(mapping.keys(), key=len, reverse=True)))
    )

    def replace_func(match):
        text = match.group(1)
        all_ = ""
        for c in text:
            all_ += _fixup_literal(ord(c), False, mapping)
        if len(text) > 1:
            multi = "".join(mapping[text])
            if len(multi) > 1:
                multi = f"[{re_escape(multi)}]"
            else:
                multi = re_escape(multi)
            return f"(?:{all_}|{multi})"
        return all_

    new = ""
    pos = 0
    for match in re.finditer(reg, u):
        new += re_escape(u[pos : match.start()])
        new += replace_func(match)
        pos = match.end()
    new += re_escape(u[pos:])

    return new


def _fixup_not_literal(literal, mapping):
    u = chr(literal)
    return "[^{}]".format("".join(re_escape(u + "".join(mapping.get(u, [])))))


def _fixup_range(start, end, mapping):
    extra = []
    for i in range(start, end + 1):
        u = chr(i)
        if u in mapping:
            extra.append(re_escape("".join(mapping[u])))

    start = re_escape(chr(start))
    end = re_escape(chr(end))
    return "{}{}-{}".format("".join(extra), start, end)


def _merge_literals(pattern):
    done = []
    current = []

    for op, av in pattern:
        op = str(op).lower()
        if op == "literal":
            current.append(av)
        else:
            if current:
                done.append(("literals", tuple(current)))
                current = []
            done.append((op, av))
    if current:
        done.append(("literals", tuple(current)))

    return done


def _construct_in(pattern, mapping):
    negate = False
    parts = []
    for op, av in _merge_literals(pattern):
        op = str(op).lower()

        if op == "range":
            start, end = av
            parts.append(_fixup_range(start, end, mapping))
        elif op == "literals":
            expanded = []
            for c in av:
                v = _fixup_literal(c, True, mapping)
                if v not in expanded:
                    expanded.append(v)
            parts.extend(expanded)
        elif op == "negate":
            negate = True
        elif op == "category":
            av = str(av).lower()
            cats = {
                "category_word": "\\w",
                "category_not_word": "\\W",
                "category_digit": "\\d",
                "category_not_digit": "\\D",
                "category_space": "\\s",
                "category_not_space": "\\S",
            }
            try:
                parts.append(cats[av])
            except KeyError as e:
                raise NotImplementedError(av) from e
        else:
            raise NotImplementedError(op)

    return "[{}{}]".format("^" if negate else "", "".join(parts))


def _construct_regexp(
    pattern: sre_parse.SubPattern, mapping: dict[str, list[str]], parent=""
) -> str:
    """Raises NotImplementedError"""

    parts = []

    for op, av in _merge_literals(pattern):
        op = str(op).lower()

        assert op != "literal"

        if op == "not_literal":
            parts.append(_fixup_not_literal(av, mapping))
        elif op == "literals":
            parts.append(_fixup_literal_list(av, mapping))
        elif op == "category":
            av = str(av).lower()
            cats = {
                "category_word": "\\w",
                "category_not_word": "\\W",
                "category_digit": "\\d",
                "category_not_digit": "\\D",
                "category_space": "\\s",
                "category_not_space": "\\S",
            }
            try:
                parts.append(cats[av])
            except KeyError as e:
                raise NotImplementedError(av) from e
        elif op == "any":
            parts.append(".")
        elif op == "in":
            parts.append(_construct_in(av, mapping))
        elif op == "max_repeat" or op == "min_repeat":
            min_, max_, pad = av
            pad = _construct_regexp(pad, mapping)
            if min_ == 1 and max_ == sre_constants.MAXREPEAT:
                parts.append(f"{pad}+")
            elif min_ == 0 and max_ == sre_constants.MAXREPEAT:
                parts.append(f"{pad}*")
            elif min_ == 0 and max_ == 1:
                parts.append(f"{pad}?")
            else:
                parts.append("%s{%d,%d}" % (pad, min_, max_))
            if op == "min_repeat":
                parts[-1] = parts[-1] + "?"
        elif op == "at":
            av = str(av).lower()
            ats = {
                "at_beginning": "^",
                "at_end": "$",
                "at_beginning_string": "\\A",
                "at_boundary": "\\b",
                "at_non_boundary": "\\B",
                "at_end_string": "\\Z",
            }
            try:
                parts.append(ats[av])
            except KeyError as e:
                raise NotImplementedError(av) from e
        elif op == "subpattern":
            # Python 3.6 extended this
            # https://bugs.python.org/issue433028
            if len(av) == 4:
                if av[1:3] == (0, 0):
                    av = [av[0], av[-1]]
                else:
                    raise NotImplementedError(op, av)
            group, pad = av
            pad = _construct_regexp(pad, mapping, parent=op)
            if group is None:
                parts.append(f"(?:{pad})")
            else:
                parts.append(f"({pad})")
        elif op == "assert":
            direction, pad = av
            pad = _construct_regexp(pad, mapping)
            if direction == 1:
                parts.append(f"(?={pad})")
            elif direction == -1:
                parts.append(f"(?<={pad})")
            else:
                raise NotImplementedError(direction)
        elif op == "assert_not":
            direction, pad = av
            pad = _construct_regexp(pad, mapping)
            if direction == 1:
                parts.append(f"(?!{pad})")
            elif direction == -1:
                parts.append(f"(?<!{pad})")
            else:
                raise NotImplementedError(direction)
        elif op == "branch":
            dummy, branches = av
            branches = (_construct_regexp(b, mapping) for b in branches)
            pad = "|".join(branches)
            if parent != "subpattern":
                parts.append(f"(?:{pad})")
            else:
                parts.append(pad)
        else:
            raise NotImplementedError(op)

    return "".join(parts)


def re_replace_literals(text: str, mapping: dict[str, list[str]]) -> str:
    """Raises NotImplementedError or re.error"""

    assert isinstance(text, str)

    pattern = sre_parse.parse(text)
    return _construct_regexp(pattern, mapping)


def re_add_variants(text: str) -> str:
    """Will replace all occurrences of ascii chars
    by a bracket expression containing the character and all its
    variants with a diacritic mark.

    "föhn" -> "[fḟ]ö[hĥȟḣḥḧḩḫẖ][nñńņňǹṅṇṉṋ]"

    In case the passed in regex is invalid raises re.error.

    Supports all regexp except ones with group references. In
    case something is not supported NotImplementedError gets raised.
    """

    assert isinstance(text, str)

    text = unicodedata.normalize("NFC", text)
    return re_replace_literals(text, get_replacement_mapping())


def compile(
    pattern: str, ignore_case: bool = True, dot_all: bool = False, asym: bool = False
) -> Callable[[str], bool]:
    """
    Args:
        pattern (str): a unicode regex
        ignore_case (bool): if case shouuld be ignored when matching
        dot_all (bool): if "." should match newlines
        asym (bool): if ascii should match similar looking unicode chars
    Returns:
        A callable which will return True if the pattern is contained in
        the passed text.
    Raises:
        ValueError: In case the regex is invalid
    """

    assert isinstance(pattern, str)

    pattern = unicodedata.normalize("NFC", pattern)

    if asym:
        try:
            pattern = re_add_variants(pattern)
        except NotImplementedError:
            # too complex, just skip this step
            print_d(f"regex not supported: {pattern}")
        except re.error as e:
            raise ValueError(e) from e

    mods = re.MULTILINE | re.UNICODE
    if ignore_case:
        mods |= re.IGNORECASE
    if dot_all:
        mods |= re.DOTALL

    try:
        reg = re.compile(pattern, mods)
    except re.error as e:
        raise ValueError(e) from e
    normalize = unicodedata.normalize

    def search(text: str):
        return bool(reg.search(normalize("NFC", text)))

    return search



# Copyright 2012-2020 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


import json
from collections import namedtuple

from quodlibet.util.dprint import print_d, print_w
from quodlibet.util.misc import total_ordering


@total_ordering
class JSONObject:
    """
    Base class for simple, named data objects
    that can be edited and persisted as JSON.
    """

    NAME = ""

    # The format for JSONObject.
    Field = namedtuple("Field", ["human_name", "doc"])
    EMPTY_FIELD = Field(None, None)

    # Override this to specify a set of field names,
    # or a dict of field: FieldData
    # Must include "name" if dict is specified.
    FIELDS: dict[str, Field] = {}

    @classmethod
    def _should_store(cls, field_name):
        """Decides if a field should be stored"""
        return not field_name.startswith("_")

    def __init__(self, name):
        if not name:
            raise ValueError(f"{type(self).__name__} objects must be named")
        self.name = str(name)

    @property
    def data(self):
        """A list of tuples of the persisted key:values in this class"""
        if self.FIELDS:
            return [
                (k, self.__getattribute__(k) if hasattr(self, k) else None)
                for k in self.FIELDS
            ]
        print_d(f"No order specified for class {type(self).__name__}")
        return {k: v for k, v in self.__dict__.items() if self._should_store(k)}

    def field(self, name):
        """Returns the Field metadata of field `name` if available,
        or a null / empty one"""
        if isinstance(self.FIELDS, dict):
            return self.FIELDS.get(name, self.EMPTY_FIELD)
        return None

    @property
    def json(self):
        return json.dumps(dict(self.data))

    def __eq__(self, other):
        return self.data == other.data

    def __lt__(self, other):
        return self.data < other.data

    def __str__(self):
        return f"<{self.__class__.__name__} '{self.name}' with {self.json}>"


class JSONObjectDict(dict):
    """A dict wrapper to persist named `JSONObject`-style objects"""

    def __init__(self, item_cls=JSONObject):
        self.Item = item_cls
        dict.__init__(self)

    @classmethod
    def from_json(cls, item_cls, json_str):
        """
        Factory method for building from an input string,
        a JSON map of {item_name1: {key:value, key2:value2...}, item_name2:...}
        """
        new = cls(item_cls)

        try:
            data = json.loads(json_str)
        except ValueError:
            print_w(f"Broken JSON: {json_str}")
        else:
            for name, blob in data.items():
                try:
                    new[name] = item_cls(**blob)
                except TypeError as e:
                    msg = f"Couldn't instantiate {item_cls.__name__} from JSON ({e})"
                    raise OSError(msg) from e
        return new

    @classmethod
    def from_list(cls, json_objects, raise_errors=True):
        new = cls()
        for j in json_objects:
            if not isinstance(j, JSONObject):
                msg = (
                    f"Incorrect type ({j.__class__.__name__}) found in list of objects"
                )
                if raise_errors:
                    raise TypeError(msg)
                print_d(msg)
            else:
                if not j.name and raise_errors:
                    raise ValueError(f"Null key for {cls.__name__} object {j}")
                if j.name in new:
                    print_w(f"Duplicate {cls.__name__} found: '{j.name}'. Removing…")
                new[j.name] = j
        return new

    def save(self, filename=None):
        """
        Takes a list of `JSONObject` objects and returns
        the data serialised as a JSON string,
        also writing (prettily) to file `filename` if specified.
        """
        print_d(f"Saving {len(self):d} {self.Item.__name__}(s) to JSON..")
        try:
            obj_dict = {o.name: dict(o.data) for o in self.values()}
        except AttributeError:
            raise
        json_str = json.dumps(obj_dict, indent=4)
        json_str = json_str.encode("utf-8")
        if filename:
            try:
                with open(filename, "wb") as f:
                    f.write(json_str)
            except OSError as e:
                print_w(
                    f"Couldn't write JSON for {type(self).__name__} object(s) ({e})"
                )
        return json_str



# Copyright 2018 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import enum
import ctypes

from gi.repository import GLib, GObject


try:
    h = ctypes.cdll.LoadLibrary("libSM.so.6")
except OSError as e:
    raise ImportError(e) from e


@ctypes.POINTER
class SmcConn(ctypes.Structure):
    pass


SmPointer = ctypes.c_void_p
Bool = ctypes.c_int

SmcSaveYourselfProc = ctypes.CFUNCTYPE(
    None, SmcConn, SmPointer, ctypes.c_int, Bool, ctypes.c_int, Bool
)
SmcDieProc = ctypes.CFUNCTYPE(None, SmcConn, SmPointer)
SmcSaveCompleteProc = ctypes.CFUNCTYPE(None, SmcConn, SmPointer)
SmcShutdownCancelledProc = ctypes.CFUNCTYPE(None, SmcConn, SmPointer)


class save_yourself(ctypes.Structure):
    _fields_ = [
        ("callback", SmcSaveYourselfProc),
        ("client_data", SmPointer),
    ]


class die(ctypes.Structure):
    _fields_ = [
        ("callback", SmcDieProc),
        ("client_data", SmPointer),
    ]


class save_complete(ctypes.Structure):
    _fields_ = [
        ("callback", SmcSaveCompleteProc),
        ("client_data", SmPointer),
    ]


class shutdown_cancelled(ctypes.Structure):
    _fields_ = [
        ("callback", SmcShutdownCancelledProc),
        ("client_data", SmPointer),
    ]


class SmcCallbacks(ctypes.Structure):
    _fields_ = [
        ("save_yourself", save_yourself),
        ("die", die),
        ("save_complete", save_complete),
        ("shutdown_cancelled", shutdown_cancelled),
    ]


SmProtoMajor = 1
SmProtoMinor = 0

SmcSaveYourselfProcMask = 1 << 0
SmcDieProcMask = 1 << 1
SmcSaveCompleteProcMask = 1 << 2
SmcShutdownCancelledProcMask = 1 << 3

SmcCloseStatus = ctypes.c_int
SmcClosedNow = 0
SmcClosedASAP = 1
SmcConnectionInUse = 2

SmcOpenConnection = h.SmcOpenConnection
SmcOpenConnection.argtypes = [
    ctypes.c_char_p,
    SmPointer,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_ulong,
    ctypes.POINTER(SmcCallbacks),
    ctypes.c_char_p,
    ctypes.POINTER(ctypes.c_char_p),
    ctypes.c_int,
    ctypes.c_char_p,
]
SmcOpenConnection.restype = SmcConn

SmcCloseConnection = h.SmcCloseConnection
SmcCloseConnection.argtypes = [SmcConn, ctypes.c_int, ctypes.POINTER(ctypes.c_char_p)]
SmcCloseConnection.restype = SmcCloseStatus

SmcSaveYourselfDone = h.SmcSaveYourselfDone
SmcSaveYourselfDone.argtypes = [SmcConn, Bool]
SmcSaveYourselfDone.restype = None


@ctypes.POINTER
class IceConn(ctypes.Structure):
    pass


IcePointer = ctypes.c_void_p
IceWatchProc = ctypes.CFUNCTYPE(
    None, IceConn, IcePointer, Bool, ctypes.POINTER(IcePointer)
)

Status = ctypes.c_int

IceAddConnectionWatch = h.IceAddConnectionWatch
IceAddConnectionWatch.argtypes = [IceWatchProc, IcePointer]
IceAddConnectionWatch.restype = Status

IceRemoveConnectionWatch = h.IceRemoveConnectionWatch
IceRemoveConnectionWatch.argtypes = [IceWatchProc, IcePointer]
IceRemoveConnectionWatch.restype = None

IceConnectionNumber = h.IceConnectionNumber
IceConnectionNumber.argtypes = [IceConn]
IceConnectionNumber.restype = ctypes.c_int

IceProcessMessagesStatus = ctypes.c_int
IceProcessMessagesSuccess = 0
IceProcessMessagesIOError = 1
IceProcessMessagesConnectionClosed = 2


@ctypes.POINTER
class FIXMEPtr(ctypes.Structure):
    pass


IceProcessMessages = h.IceProcessMessages
IceProcessMessages.argtypes = [IceConn, FIXMEPtr, FIXMEPtr]
IceProcessMessages.restype = IceProcessMessagesStatus

IceSetShutdownNegotiation = h.IceSetShutdownNegotiation
IceSetShutdownNegotiation.argtypes = [IceConn, Bool]
IceSetShutdownNegotiation.restype = None

IceCloseStatus = ctypes.c_int

IceCloseConnection = h.IceCloseConnection
IceCloseConnection.argtypes = [IceConn]
IceCloseConnection.restype = IceCloseStatus


class SaveType(enum.IntEnum):
    GLOBAL = 0
    LOCAL = 1
    BOTH = 2


class InteractStyle(enum.IntEnum):
    NONE = 0
    ERRORS = 1
    ANY = 2


class XSMPError(Exception):
    pass


class XSMPSource:
    """Dispatches SM messages in the glib mainloop"""

    def __init__(self):
        self._watch_id = None
        self._watch_proc = None

    def open(self):
        if self._watch_proc is not None:
            raise XSMPError("already open")

        @IceWatchProc
        def watch_proc(conn, client_data, opening, watch_data):
            if opening:
                fd = IceConnectionNumber(conn)
                channel = GLib.IOChannel.unix_new(fd)
                self._watch_id = GLib.io_add_watch(
                    channel,
                    GLib.PRIORITY_DEFAULT,
                    (GLib.IOCondition.ERR | GLib.IOCondition.HUP | GLib.IOCondition.IN),
                    self._process_func,
                    conn,
                )
            else:
                if self._watch_id is not None:
                    GObject.source_remove(self._watch_id)
                    self._watch_id = None

        self._watch_proc = watch_proc
        status = IceAddConnectionWatch(watch_proc, None)
        if status == 0:
            raise XSMPError("IceAddConnectionWatch failed with %d" % status)

    def close(self):
        if self._watch_proc is not None:
            IceRemoveConnectionWatch(self._watch_proc, None)
            self._watch_proc = None

        if self._watch_id is not None:
            GObject.source_remove(self._watch_id)
            self._watch_id = None

    def _process_func(self, channel, condition, conn):
        status = IceProcessMessages(conn, None, None)
        if status != IceProcessMessagesSuccess:
            if status != IceProcessMessagesConnectionClosed:
                IceCloseConnection(conn)
            self._watch_id = None
            return False
        return True


class XSMPClient(GObject.Object):
    __gsignals__ = {
        "save-yourself": (
            GObject.SignalFlags.RUN_LAST,
            None,
            (object, object, object, object),
        ),
        "die": (GObject.SignalFlags.RUN_LAST, None, ()),
        "save-complete": (GObject.SignalFlags.RUN_LAST, None, ()),
        "shutdown-cancelled": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self):
        super().__init__()

        self._source = None
        self._callbacks = SmcCallbacks()
        self._conn = None
        self._id = None

        def wrap_cb(func_type, cb):
            def c_callback(*args):
                return cb(self, func_type, *args[2:])

            return func_type(c_callback)

        self._callbacks.save_yourself.callback = wrap_cb(
            SmcSaveYourselfProc, self._on_save_yourself
        )
        self._callbacks.die.callback = wrap_cb(SmcDieProc, self._on_die)
        self._callbacks.save_complete.callback = wrap_cb(
            SmcSaveCompleteProc, self._on_save_complete
        )
        self._callbacks.shutdown_cancelled.callback = wrap_cb(
            SmcShutdownCancelledProc, self._on_shutdown_cancelled
        )

    def _on_save_yourself(
        self, conn, client_data, save_type, shutdown, interact_style, fast
    ):
        self.emit(
            "save-yourself",
            SaveType(save_type),
            bool(shutdown),
            InteractStyle(interact_style),
            bool(fast),
        )

    def _on_die(self, conn, client_data):
        self.emit("die")

    def _on_save_complete(self, conn, client_data):
        self.emit("save-complete")

    def _on_shutdown_cancelled(self, conn, client_data):
        self.emit("shutdown-cancelled")

    @property
    def client_id(self):
        if self._conn is None:
            raise XSMPError("connection closed")
        return self._id

    def open(self):
        if self._conn is not None:
            raise XSMPError("connection already open")

        self._source = XSMPSource()
        self._source.open()

        error_string = ctypes.create_string_buffer(250)
        id_ = ctypes.c_char_p()
        self._conn = SmcOpenConnection(
            None,
            None,
            SmProtoMajor,
            SmProtoMinor,
            (
                SmcDieProcMask
                | SmcSaveCompleteProcMask
                | SmcSaveYourselfProcMask
                | SmcShutdownCancelledProcMask
            ),
            ctypes.byref(self._callbacks),
            None,
            ctypes.byref(id_),
            len(error_string),
            error_string,
        )
        # null ptr still returns an object, but its falsy
        if not self._conn:
            self._conn = None

        if self._conn is None:
            self._conn = None
            self._source.close()
            self._source = None
            raise XSMPError(
                "open failed: {!r}".format(error_string.value.decode("utf-8"))
            )

        # FIXME: id_ should be freed with free()
        self._id = id_.value.decode("utf-8")

    def save_yourself_done(self, success):
        if self._conn is None:
            raise XSMPError("connection closed")
        SmcSaveYourselfDone(self._conn, success)

    def close(self):
        if self._conn is not None:
            SmcCloseConnection(self._conn, 0, None)
            self._conn = None
        if self._source is not None:
            self._source.close()
            self._source = None

    def __del__(self):
        self.close()
        self._callbacks = None



# Copyright 2010,2012 Christoph Reiter <reiter.christoph@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import dbus


class MPRISObject(dbus.service.Object):
    def paused(self):
        pass

    def unpaused(self):
        pass

    def song_started(self, song):
        pass

    def song_ended(self, song, skipped):
        pass



# Copyright 2004-2005 Joe Wreschnig, Michael Urman, Iñigo Serna
#           2012 Christoph Reiter
#           2016-23 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from typing import cast

from quodlibet import util
from quodlibet.util.importhelper import load_dir_modules

from ._base import Browser

browsers: list[type[Browser]] = []

BrowserName = str
BrowserKey = int | BrowserName
"""Either the index or the string key of the browser"""

default = None


def init() -> None:
    """Import all browsers from this package and from the user directory.

    After this is called the global `browsers` list will contain all
    classes sorted by priority.

    Can be called multiple times.
    """

    global browsers, default

    # ignore double init (for the test suite)
    if browsers:
        return

    this_dir = util.get_module_dir()
    modules = load_dir_modules(this_dir, package=__package__)

    for browser in modules:
        try:
            browsers.extend(browser.browsers)
        except AttributeError:
            util.print_w(f"{browser.__name__!r} doesn't contain any browsers.")

    def is_browser(klass):
        return isinstance(klass, type) and issubclass(klass, Browser)

    browsers = list(filter(is_browser, browsers))

    if not browsers:
        raise SystemExit("No browsers found!")

    browsers.sort(key=lambda cls: cls.priority)

    try:
        default = get("SearchBar")
    except ValueError as e:
        raise SystemExit("Default browser not found!") from e


def name(browser: Browser) -> BrowserName:
    """Return the name of the browser"""

    return browser.keys[0]


def get(i: BrowserKey) -> type[Browser]:
    """Return a constructor for a browser, either given by number, a string
    of the number, or the name.

    Raises ValueError if the lookup fails.
    """

    try:
        return browsers[int(i)]
    except (IndexError, ValueError, TypeError):
        try:
            return get(index(cast(str, i)))
        except IndexError as e:
            # ValueError will fall through
            raise ValueError(f"{i!r} not found") from e


def index(name: BrowserName) -> int:
    """Return the index of a browser given its name.

    Raises ValueError if the lookup fails.
    """

    name = name.lower()
    for j, browser in enumerate(browsers):
        keys = [k.lower() for k in browser.keys]
        if name in keys:
            return j

    all_keys = (k.lower() for b in browsers for k in b.keys)
    raise ValueError(f"{name!r} browser not found. Try: {' | '.join(all_keys)}")



# Copyright 2004-2005 Joe Wreschnig, Michael Urman, Iñigo Serna
#           2016-2020 Nick Boultbee
#                2022 Jej@github
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk, Pango, GObject, Gdk

from quodlibet import _
from quodlibet import config
from quodlibet import const
from quodlibet import qltk
from quodlibet import util
from quodlibet.plugins import PluginManager, plugin_enabled, Plugin
from quodlibet.plugins.cover import CoverSourcePlugin
from quodlibet.plugins.editing import EditTagsPlugin, RenameFilesPlugin
from quodlibet.plugins.events import EventPlugin
from quodlibet.plugins.gstelement import GStreamerPlugin
from quodlibet.plugins.playlist import PlaylistPlugin
from quodlibet.plugins.playorder import PlayOrderPlugin
from quodlibet.plugins.query import QueryPlugin
from quodlibet.plugins.songsmenu import SongsMenuPlugin
from quodlibet.qltk import Icons, is_accel, show_uri
from quodlibet.qltk.entry import UndoEntry
from quodlibet.qltk.models import ObjectStore, ObjectModelFilter
from quodlibet.qltk.views import HintedTreeView
from quodlibet.qltk.window import UniqueWindow, PersistentWindowMixin
from quodlibet.qltk.x import Align, Paned, Button, ScrolledWindow
from quodlibet.util import connect_obj
from quodlibet.util.string.filter import remove_diacritics

PLUGIN_CATEGORIES = {
    _("Songs"): SongsMenuPlugin,
    _("Playlists"): PlaylistPlugin,
    _("Events"): EventPlugin,
    _("Play Order"): PlayOrderPlugin,
    _("Editing"): EditTagsPlugin,
    _("Renaming"): RenameFilesPlugin,
    _("Querying"): QueryPlugin,
    _("Effects"): GStreamerPlugin,
    _("Covers"): CoverSourcePlugin,
}


def category_of(plugin: Plugin) -> str:
    try:
        return next(
            cat for cat, cls in PLUGIN_CATEGORIES.items() if issubclass(plugin.cls, cls)
        )
    except StopIteration:
        return _("Unknown")


class UndoSearchEntry(Gtk.SearchEntry, UndoEntry):
    pass


class PluginErrorWindow(UniqueWindow):
    def __init__(self, parent, failures):
        if self.is_not_unique():
            return
        super().__init__()

        self.set_title(_("Plugin Errors"))
        self.set_border_width(6)
        self.set_transient_for(parent)
        self.set_default_size(520, 300)

        scrolledwin = Gtk.ScrolledWindow()
        vbox = Gtk.VBox(spacing=6)
        vbox.set_border_width(6)
        scrolledwin.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolledwin.add_with_viewport(vbox)

        keys = failures.keys()
        show_expanded = len(keys) <= 3
        for key in sorted(keys):
            expander = Gtk.Expander(label=util.bold(key))
            expander.set_use_markup(True)
            if show_expanded:
                expander.set_expanded(True)

            # second line is always the __rescan line; don't show it
            message = failures[key][0:1] + failures[key][3:]
            failure = Gtk.Label(label="".join(message).strip())
            failure.set_alignment(0, 0)
            failure.set_padding(12, 6)
            failure.set_selectable(True)
            failure.set_line_wrap(True)

            vbox.pack_start(expander, False, True, 0)
            expander.add(failure)

        self.use_header_bar()

        if not self.has_close_button():
            vbox2 = Gtk.VBox(spacing=12)
            close = Button(_("_Close"), Icons.WINDOW_CLOSE)
            close.connect("clicked", lambda *x: self.destroy())
            b = Gtk.HButtonBox()
            b.set_layout(Gtk.ButtonBoxStyle.END)
            b.pack_start(close, True, True, 0)
            vbox2.pack_start(scrolledwin, True, True, 0)
            vbox2.pack_start(b, False, True, 0)
            self.add(vbox2)
            close.grab_focus()
        else:
            self.add(scrolledwin)

        self.get_child().show_all()


class EnabledType:
    TAG, ALL, NO, DIS, EN, SEP = range(6)


class PluginEnabledFilterCombo(Gtk.ComboBox):
    def __init__(self):
        combo_store = Gtk.ListStore(str, int)
        super().__init__(model=combo_store)

        cell = Gtk.CellRendererText()
        cell.props.ellipsize = Pango.EllipsizeMode.END
        self.pack_start(cell, True)
        self.add_attribute(cell, "text", 0)

        def combo_sep(model, iter_, data):
            return model[iter_][1] == EnabledType.SEP

        self.set_row_separator_func(combo_sep, None)

    def refill(self, tags, no_tags):
        """Fill with a sequence of tags.
        If no_tags is true display display the extra category for it.
        """

        active = max(self.get_active(), 0)
        combo_store = self.get_model()
        combo_store.clear()
        combo_store.append([_("Any state"), EnabledType.ALL])
        combo_store.append(["", EnabledType.SEP])
        combo_store.append([_("Enabled"), EnabledType.EN])
        combo_store.append([_("Disabled"), EnabledType.DIS])
        if tags:
            combo_store.append(["", EnabledType.SEP])
            for tag in sorted(tags):
                combo_store.append([tag, EnabledType.TAG])
            if no_tags:
                combo_store.append([_("No category"), EnabledType.NO])
        self.set_active(active)

    def get_active_row(self):
        iter_ = self.get_active_iter()
        if iter_:
            model = self.get_model()
            return list(model[iter_])
        return None


class PluginTypeFilterCombo(Gtk.ComboBox):
    def __init__(self):
        combo_store = Gtk.ListStore(str, object)
        super().__init__(model=combo_store)

        cell = Gtk.CellRendererText()
        cell.props.ellipsize = Pango.EllipsizeMode.END
        self.pack_start(cell, True)
        self.add_attribute(cell, "text", 0)

        def combo_sep(model, iter_, data):
            return model[iter_][1] is None

        self.set_row_separator_func(combo_sep, None)
        self.__refill()

    def __refill(self):
        """Fill with plugin types"""

        active = max(self.get_active(), 0)
        combo_store = self.get_model()
        combo_store.clear()
        combo_store.append([_("Any category"), object])
        combo_store.append(["", None])
        for name, cls in PLUGIN_CATEGORIES.items():
            combo_store.append([name, cls])

        self.set_active(active)

    def get_active_type(self):
        iter_ = self.get_active_iter()
        if iter_:
            model = self.get_model()
            return model[iter_][1]
        return None


class PluginListView(HintedTreeView):
    __gsignals__ = {
        # model, iter, enabled
        "plugin-toggled": (GObject.SignalFlags.RUN_LAST, None, (object, object, bool))
    }

    def __init__(self):
        super().__init__()
        self.set_headers_visible(False)

        render = Gtk.CellRendererToggle()
        render.set_padding(6, 3)

        def cell_data(col, render, model, iter_, data):
            plugin = model.get_value(iter_)
            pm = PluginManager.instance
            render.set_activatable(plugin.can_enable)
            # If it can't be enabled because it's an always-on kinda thing,
            # show it as enabled so it doesn't look broken.
            render.set_active(pm.enabled(plugin) or not plugin.can_enable)

        render.connect("toggled", self.__toggled)
        column = Gtk.TreeViewColumn("enabled", render)
        column.set_cell_data_func(render, cell_data)
        self.append_column(column)

        render = Gtk.CellRendererPixbuf()
        render.set_padding(1, 1)

        def cell_data2(col, render, model, iter_, data):
            plugin = model.get_value(iter_)
            icon = plugin.icon or Icons.SYSTEM_RUN
            render.set_property("icon-name", icon)
            render.set_property("stock-size", Gtk.IconSize.LARGE_TOOLBAR)

        column = Gtk.TreeViewColumn("image", render)
        column.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        column.set_cell_data_func(render, cell_data2)
        self.append_column(column)

        render = Gtk.CellRendererText()
        render.set_property("ellipsize", Pango.EllipsizeMode.END)
        render.set_property("xalign", 0.0)
        render.set_padding(6, 6)
        column = Gtk.TreeViewColumn("name", render)

        def cell_data3(col, render, model, iter_, data):
            plugin = model.get_value(iter_)
            render.set_property("text", plugin.name)

        column.set_cell_data_func(render, cell_data3)
        column.set_expand(True)
        self.append_column(column)

    def do_key_press_event(self, event):
        if is_accel(event, "space", "KP_Space"):
            selection = self.get_selection()
            fmodel, fiter = selection.get_selected()
            plugin = fmodel.get_value(fiter)
            if plugin.can_enable:
                self._emit_toggled(fmodel.get_path(fiter), not plugin_enabled(plugin))
            self.get_model().iter_changed(fiter)
        else:
            Gtk.TreeView.do_key_press_event(self, event)

    def __toggled(self, render, path):
        render.set_active(not render.get_active())
        self._emit_toggled(path, render.get_active())

    def _emit_toggled(self, path, value):
        model = self.get_model()
        iter_ = model.get_iter(path)
        self.emit("plugin-toggled", model, iter_, value)

    def select_by_plugin_id(self, plugin_id):
        def restore_sel(row):
            return row[0].id == plugin_id

        if not self.select_by_func(restore_sel, one=True):
            self.set_cursor((0,))

    def refill(self, plugins):
        selection = self.get_selection()

        fmodel, fiter = selection.get_selected()
        model = fmodel.get_model()

        # get the ID of the selected plugin
        selected = None
        if fiter:
            plugin = fmodel.get_value(fiter)
            selected = plugin.id

        model.clear()

        for plugin in sorted(plugins, key=lambda x: x.name):
            it = model.append(row=[plugin])
            if plugin.id == selected:
                ok, fit = fmodel.convert_child_iter_to_iter(it)
                selection.select_iter(fit)


class PluginPreferencesContainer(Gtk.VBox):
    def __init__(self):
        super().__init__(spacing=12)

        self.desc = desc = Gtk.Label()
        desc.set_line_wrap(True)
        # Ensure a reasonable minimum height request for long descriptions
        desc.set_width_chars(30)
        desc.set_alignment(0, 0.5)
        desc.set_selectable(True)
        self.pack_start(desc, False, True, 0)

        self.prefs = prefs = Gtk.Frame()
        prefs.set_shadow_type(Gtk.ShadowType.NONE)
        self.pack_start(prefs, False, True, 0)

    def set_no_plugins(self):
        self.set_plugin(None)
        self.desc.set_text(_("No plugins found."))

    def set_plugin(self, plugin):
        label = self.desc

        if plugin is None:
            label.set_markup("")
        else:
            name = util.escape(plugin.name)
            category = category_of(plugin).lower()
            text = (
                f"<big><b>{name}</b> "
                f"<span alpha='40%'> – {category}</span>"
                f"</big>"
            )
            markup = plugin.description_markup or _("(undocumented)")
            if markup:
                text += f"<span font='4'>\n\n</span>{markup}"
            label.set_markup(text)
            label.connect("activate-link", show_uri)

        frame = self.prefs

        if frame.get_child():
            frame.get_child().destroy()

        if plugin is None:
            frame.hide()
        else:
            instance_or_cls = plugin.get_instance() or plugin.cls

            if plugin and hasattr(instance_or_cls, "PluginPreferences"):
                try:
                    prefs = instance_or_cls.PluginPreferences(self)
                except Exception:
                    util.print_exc()
                    frame.hide()
                else:
                    if isinstance(prefs, Gtk.Window):
                        b = Button(_("_Preferences"), Icons.PREFERENCES_SYSTEM)
                        connect_obj(b, "clicked", Gtk.Window.show, prefs)
                        connect_obj(b, "destroy", Gtk.Window.destroy, prefs)
                        frame.add(b)
                        frame.get_child().set_border_width(6)
                    else:
                        frame.add(prefs)
                    frame.show_all()


class PluginWindow(UniqueWindow, PersistentWindowMixin):
    def __init__(self, parent=None):
        if self.is_not_unique():
            return
        on_top = config.getboolean("settings", "plugins_window_on_top", False)
        super().__init__(dialog=on_top)
        self.set_title(_("Plugins"))
        self.set_default_size(750, 500)
        if parent and on_top:
            self.set_transient_for(parent)
        self.set_type_hint(Gdk.WindowTypeHint.NORMAL)
        self.enable_window_tracking("plugin_prefs")

        model = ObjectStore()
        filter_model = ObjectModelFilter(child_model=model)

        self._list_view = plv = PluginListView()
        plv.set_model(filter_model)
        plv.set_rules_hint(True)

        plv.connect("plugin-toggled", self.__plugin_toggled)
        sw = ScrolledWindow()
        sw.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.ALWAYS)
        sw.add(plv)
        sw.set_shadow_type(Gtk.ShadowType.IN)

        fb = Gtk.HBox(spacing=6)

        enabled_combo = PluginEnabledFilterCombo()
        enabled_combo.connect("changed", lambda s: filter_model.refilter())
        enabled_combo.set_tooltip_text(_("Filter by plugin state / tag"))
        fb.pack_start(enabled_combo, True, True, 0)
        self._enabled_combo = enabled_combo

        type_combo = PluginTypeFilterCombo()
        type_combo.connect("changed", lambda s: filter_model.refilter())
        type_combo.set_tooltip_text(_("Filter by plugin type"))
        fb.pack_start(type_combo, True, True, 0)
        self._type_combo = type_combo

        self._filter_entry = fe = UndoSearchEntry()
        fe.set_tooltip_text(_("Filter by plugin name or description"))
        fe.connect("changed", lambda s: filter_model.refilter())

        errors = qltk.Button(_("Show _Errors"), Icons.DIALOG_WARNING)
        errors.set_focus_on_click(False)
        errors.connect("clicked", self.__show_errors)
        errors.show()
        errors = Align(errors, top=6, bottom=6)
        errors.set_no_show_all(True)
        bbox = Gtk.VBox()
        bbox.pack_start(errors, True, True, 0)

        pref_box = PluginPreferencesContainer()

        if const.DEBUG:
            refresh = qltk.Button(_("_Refresh"), Icons.VIEW_REFRESH)
            refresh.set_focus_on_click(False)
            refresh.connect(
                "clicked", self.__refresh, plv, pref_box, errors, enabled_combo
            )
            bbox.pack_start(refresh, True, True, 0)

        filter_box = Gtk.VBox(spacing=6)
        filter_box.pack_start(fb, False, True, 0)
        filter_box.pack_start(fe, False, True, 0)

        vbox = Gtk.VBox()
        vbox.pack_start(Align(filter_box, border=6, right=-6), False, False, 0)
        vbox.pack_start(sw, True, True, 0)
        vbox.pack_start(Align(bbox, left=3, right=3, top=0), False, False, 3)
        paned = Paned()
        paned.pack1(vbox, False, False)

        close = qltk.Button(_("_Close"), Icons.WINDOW_CLOSE)
        close.connect("clicked", lambda *x: self.destroy())
        bb_align = Align(halign=Gtk.Align.END, valign=Gtk.Align.END)
        bb = Gtk.HButtonBox()
        bb.set_layout(Gtk.ButtonBoxStyle.END)
        bb.pack_start(close, True, True, 0)
        bb_align.add(bb)

        selection = plv.get_selection()
        selection.connect("changed", self.__selection_changed, pref_box)
        selection.emit("changed")

        right_box = Gtk.VBox()
        right_box.pack_start(pref_box, True, True, 0)
        if not self.has_close_button():
            right_box.pack_start(bb_align, True, True, 0)

        align = Align(right_box, left=6, right=15, top=12, bottom=3)
        paned.pack2(align, True, False)
        paned.set_position(290)

        self.add(paned)

        self.__refill(plv, pref_box, errors, enabled_combo)

        self.connect("destroy", self.__destroy)
        filter_model.set_visible_func(self.__filter, (fe, enabled_combo, type_combo))

        self.get_child().show_all()
        fe.grab_focus()

        restore_id = config.get("memory", "plugin_selection")
        plv.select_by_plugin_id(restore_id)

    def __filter(self, model, iter_, data):
        """Filter a single row"""
        plugin = model.get_value(iter_)
        if not plugin:
            return False

        entry, state_combo, type_combo = data

        plugin_type = type_combo.get_active_type()
        if not issubclass(plugin.cls, plugin_type):
            return False

        tag_row = state_combo.get_active_row()
        if tag_row:
            plugin_tags = plugin.tags
            tag, flag = tag_row
            enabled = plugin_enabled(plugin)
            if (
                flag == EnabledType.NO
                and plugin_tags
                or flag == EnabledType.TAG
                and tag not in plugin_tags
                or flag == EnabledType.EN
                and not enabled
                or flag == EnabledType.DIS
                and enabled
            ):
                return False

        def matches(text, filter_):
            return all(
                p in remove_diacritics(text.lower()) for p in filter_.lower().split()
            )

        filter_ = remove_diacritics(entry.get_text())
        return (
            matches(plugin.name, filter_)
            or matches(plugin.id, filter_)
            or matches((plugin.description or ""), filter_)
        )

    def __destroy(self, *args):
        config.save()

    def __selection_changed(self, selection, container):
        model, iter_ = selection.get_selected()
        if not iter_:
            container.set_plugin(None)
            return

        plugin = model.get_value(iter_)
        config.set("memory", "plugin_selection", plugin.id)
        container.set_plugin(plugin)

    def unfilter(self):
        """Clears all filters applied to the list"""

        self._enabled_combo.set_active(0)
        self._type_combo.set_active(0)
        self._filter_entry.set_text("")

    def move_to(self, plugin_id):
        def selector(r):
            return r[0].id == plugin_id

        if self._list_view.select_by_func(selector):
            return True
        self.unfilter()
        return self._list_view.select_by_func(selector)

    def __plugin_toggled(self, tv, model, iter_, enabled):
        plugin = model.get_value(iter_)
        pm = PluginManager.instance
        pm.enable(plugin, enabled)
        pm.save()

        rmodel = model.get_model()
        riter = model.convert_iter_to_child_iter(iter_)
        rmodel.row_changed(rmodel.get_path(riter), riter)

    def __refill(self, view, prefs, errors, state_combo):
        pm = PluginManager.instance

        # refill plugin list
        view.refill(pm.plugins)

        # get all tags and refill tag-based (state) combobox
        tags = set()
        no_tags = False
        for plugin in pm.plugins:
            if not plugin.tags:
                no_tags = True
            tags.update(plugin.tags)

        state_combo.refill(tags, no_tags)

        if not len(pm.plugins):
            prefs.set_no_plugins()

        errors.set_visible(bool(pm.failures))

    def __refresh(self, activator, view, prefs, errors, state_combo):
        pm = PluginManager.instance
        pm.rescan()
        self.__refill(view, prefs, errors, state_combo)

    def __show_errors(self, activator):
        pm = PluginManager.instance
        window = PluginErrorWindow(self, pm.failures)
        window.show()



# Copyright 2012,2013 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import sys
from optparse import OptionParser

from quodlibet import _
from quodlibet.formats import MusicFile, AudioFileError
from quodlibet.util import print_


class CommandError(Exception):
    pass


class Command:
    """Base class for commands.

    Subclasses can override _add_options() and _execute()
    """

    NAME = ""
    DESCRIPTION = ""
    USAGE = ""
    COMMANDS: "list[type[Command]]" = []

    @classmethod
    def register(cls, cmd_cls):
        cls.COMMANDS.append(cmd_cls)
        cls.COMMANDS.sort(key=lambda c: c.NAME)

    def __init__(self, main_cmd, options=None):
        self._main_cmd = main_cmd
        usage = f"{main_cmd} {self.NAME} {self.USAGE}"
        self.__parser = OptionParser(usage=usage, description=self.DESCRIPTION)
        if options is None:
            options = self.__parser.parse_args([])[0]
        self.__options = options
        self._add_options(self.__parser)

    def _add_options(self, parser):
        """Override to add options to the parser"""

    @property
    def verbose(self):
        return self.__options.verbose

    @verbose.setter
    def verbose(self, value):
        self.__options.verbose = bool(value)

    def log(self, text):
        """Print output if --verbose was passed"""

        if self.verbose:
            return print_(text, file=sys.stderr)
        return None

    def load_song(self, path):
        """Load a song. Raises CommandError in case it fails"""

        self.log(f"Load file: {path!r}")
        song = MusicFile(path)
        if not song:
            raise CommandError(_("Failed to load file: %r") % path)
        return song

    def save_songs(self, songs):
        """Save all passed songs"""

        self.log("Saving songs...")

        for song in songs:
            try:
                song.write()
            except AudioFileError as e:
                raise CommandError(e) from e

    def _execute(self, options, args):
        """Override to execute something"""

        raise NotImplementedError

    def print_help(self, file=None):
        """Print the help information about the command"""

        if file is None:
            file = sys.stdout

        self.__parser.print_help(file=file)

    def execute(self, args):
        """Execute the command"""

        options, args = self.__parser.parse_args(args)
        self._execute(options, args)
